import sys

from metaflow.event_logger import NullEventLogger
from metaflow.sidecar import Message, MessageTypes


class DebugEventLogger(NullEventLogger):
    TYPE = "debugLogger"

    @classmethod
    def get_worker(cls):
        return DebugEventLoggerSidecar


class DebugEventLoggerSidecar(object):
    def __init__(self):
        pass

    def process_message(self, msg):
        # type: (Message) -> None
        if msg.msg_type == MessageTypes.SHUTDOWN:
            print("Debug[shutdown]: got shutdown!", file=sys.stderr)
            self._shutdown()
        elif msg.msg_type == MessageTypes.BEST_EFFORT:
            print("Debug[best_effort]: %s" % str(msg.payload), file=sys.stderr)
        elif msg.msg_type == MessageTypes.MUST_SEND:
            print("Debug[must_send]: %s" % str(msg.payload), file=sys.stderr)

    def _shutdown(self):
        sys.stderr.flush()



from metaflow import FlowSpec, step, Parameter, conda, conda_base


def get_python_version():
    """
    A convenience function to get the python version used to run this
    tutorial. This ensures that the conda environment is created with an
    available version of python.

    """
    import platform

    versions = {"2": "2.7.15", "3": "3.9.10"}
    return versions[platform.python_version_tuple()[0]]


# Use the specified version of python for this flow.
@conda_base(python=get_python_version())
class PlayListFlow(FlowSpec):
    """
    The next version of our playlist generator that adds a 'hint' parameter to
    choose a bonus movie closest to the 'hint'.

    The flow performs the following steps:

    1) Load the genre-specific statistics from the MovieStatsFlow.
    2) In parallel branches:
       - A) Build a playlist from the top films in the requested genre.
       - B) Choose a bonus movie that has the closest string edit distance to
         the user supplied hint.
    3) Join the two to create a movie playlist and display it.

    """

    genre = Parameter(
        "genre", help="Filter movies for a particular genre.", default="Sci-Fi"
    )

    hint = Parameter(
        "hint",
        help="Give a hint to the bonus movie algorithm.",
        default="Metaflow Release",
    )

    recommendations = Parameter(
        "recommendations",
        help="The number of movies recommended for the playlist.",
        default=5,
    )

    @step
    def start(self):
        """
        Use the Metaflow client to retrieve the latest successful run from our
        MovieStatsFlow and assign them as data artifacts in this flow.

        """
        # Load the analysis from the MovieStatsFlow.
        from metaflow import Flow, get_metadata

        # Print metadata provider
        print("Using metadata provider: %s" % get_metadata())

        # Load the analysis from the MovieStatsFlow.
        run = Flow("MovieStatsFlow").latest_successful_run
        print("Using analysis from '%s'" % str(run))

        # Get the dataframe from the start step before we sliced into into
        # genre-specific dataframes.
        self.dataframe = run["start"].task.data.dataframe

        # Also grab the summary statistics.
        self.genre_stats = run.data.genre_stats

        # Compute our two recommendation types in parallel.
        self.next(self.bonus_movie, self.genre_movies)

    @conda(libraries={"editdistance": "0.5.3"})
    @step
    def bonus_movie(self):
        """
        Use the user supplied 'hint' argument to choose a bonus movie that has
        the closest string edit distance to the hint.

        This step uses 'conda' to isolate the environment. Note that the
        package 'editdistance' need not be installed in your python
        environment.
        """
        import editdistance

        # Define a helper function to compute the similarity between two
        # strings.
        def _edit_distance(movie_title):
            return editdistance.eval(self.hint, movie_title)

        # Compute the distance and take the argmin to find the closest title.
        distance = [
            _edit_distance(movie_title) for movie_title in self.dataframe["movie_title"]
        ]
        index = distance.index(min(distance))
        self.bonus = (
            self.dataframe["movie_title"][index],
            self.dataframe["genres"][index],
        )

        self.next(self.join)

    @step
    def genre_movies(self):
        """
        Select the top performing movies from the use specified genre.
        """

        from random import shuffle

        # For the genre of interest, generate a potential playlist using only
        # highest gross box office titles (i.e. those in the last quartile).
        genre = self.genre.lower()
        if genre not in self.genre_stats:
            self.movies = []
        else:
            df = self.genre_stats[genre]["dataframe"]
            quartiles = self.genre_stats[genre]["quartiles"]
            self.movies = [
                df["movie_title"][i]
                for i, g in enumerate(df["gross"])
                if g >= quartiles[-1]
            ]

        # Shuffle the content.
        shuffle(self.movies)

        self.next(self.join)

    @step
    def join(self, inputs):
        """
        Join our parallel branches and merge results.

        """
        self.playlist = inputs.genre_movies.movies
        self.bonus = inputs.bonus_movie.bonus

        self.next(self.end)

    @step
    def end(self):
        """
        This step simply prints out the playlist.

        """
        # Print the playlist.
        print("Playlist for movies in genre '%s'" % self.genre)
        for pick, movie in enumerate(self.playlist, start=1):
            print("Pick %d: '%s'" % (pick, movie))
            if pick >= self.recommendations:
                break

        print("Bonus Pick: '%s' from '%s'" % (self.bonus[0], self.bonus[1]))


if __name__ == "__main__":
    PlayListFlow()



from .base_sensor import AirflowSensorDecorator
from ..airflow_utils import SensorNames
from ..exception import AirflowException
from datetime import timedelta


AIRFLOW_STATES = dict(
    QUEUED="queued",
    RUNNING="running",
    SUCCESS="success",
    SHUTDOWN="shutdown",  # External request to shut down,
    FAILED="failed",
    UP_FOR_RETRY="up_for_retry",
    UP_FOR_RESCHEDULE="up_for_reschedule",
    UPSTREAM_FAILED="upstream_failed",
    SKIPPED="skipped",
)


class ExternalTaskSensorDecorator(AirflowSensorDecorator):
    """
    The `@airflow_external_task_sensor` decorator attaches a Airflow [ExternalTaskSensor](https://airflow.apache.org/docs/apache-airflow/stable/_api/airflow/sensors/external_task/index.html#airflow.sensors.external_task.ExternalTaskSensor) before the start step of the flow.
    This decorator only works when a flow is scheduled on Airflow and is compiled using `airflow create`. More than one `@airflow_external_task_sensor` can be added as a flow decorators. Adding more than one decorator will ensure that `start` step starts only after all sensors finish.

    Parameters
    ----------
    timeout : int
        Time, in seconds before the task times out and fails. (Default: 3600)
    poke_interval : int
        Time in seconds that the job should wait in between each try. (Default: 60)
    mode : str
        How the sensor operates. Options are: { poke | reschedule }. (Default: "poke")
    exponential_backoff : bool
        allow progressive longer waits between pokes by using exponential backoff algorithm. (Default: True)
    pool : str
        the slot pool this task should run in,
        slot pools are a way to limit concurrency for certain tasks. (Default:None)
    soft_fail : bool
        Set to true to mark the task as SKIPPED on failure. (Default: False)
    name : str
        Name of the sensor on Airflow
    description : str
        Description of sensor in the Airflow UI
    external_dag_id : str
        The dag_id that contains the task you want to wait for.
    external_task_ids : List[str]
        The list of task_ids that you want to wait for.
        If None (default value) the sensor waits for the DAG. (Default: None)
    allowed_states : List[str]
        Iterable of allowed states, (Default: ['success'])
    failed_states : List[str]
        Iterable of failed or dis-allowed states. (Default: None)
    execution_delta : datetime.timedelta
        time difference with the previous execution to look at,
        the default is the same logical date as the current task or DAG. (Default: None)
    check_existence: bool
        Set to True to check if the external task exists or check if
        the DAG to wait for exists. (Default: True)
    """

    operator_type = SensorNames.EXTERNAL_TASK_SENSOR
    # Docs:
    # https://airflow.apache.org/docs/apache-airflow/stable/_api/airflow/sensors/external_task/index.html#airflow.sensors.external_task.ExternalTaskSensor
    name = "airflow_external_task_sensor"
    defaults = dict(
        **AirflowSensorDecorator.defaults,
        external_dag_id=None,
        external_task_ids=None,
        allowed_states=[AIRFLOW_STATES["SUCCESS"]],
        failed_states=None,
        execution_delta=None,
        check_existence=True,
        # We cannot add `execution_date_fn` as it requires a python callable.
        # Passing around a python callable is non-trivial since we are passing a
        # callable from metaflow-code to airflow python script. Since we cannot
        # transfer dependencies of the callable, we cannot gaurentee that the callable
        # behave exactly as the user expects
    )

    def serialize_operator_args(self):
        task_args = super().serialize_operator_args()
        if task_args["execution_delta"] is not None:
            task_args["execution_delta"] = dict(
                seconds=task_args["execution_delta"].total_seconds()
            )
        return task_args

    def validate(self):
        if self.attributes["external_dag_id"] is None:
            raise AirflowException(
                "`%s` argument of `@%s`cannot be `None`."
                % ("external_dag_id", self.name)
            )

        if type(self.attributes["allowed_states"]) == str:
            if self.attributes["allowed_states"] not in list(AIRFLOW_STATES.values()):
                raise AirflowException(
                    "`%s` is an invalid input of `%s` for `@%s`. Accepted values are %s"
                    % (
                        str(self.attributes["allowed_states"]),
                        "allowed_states",
                        self.name,
                        ", ".join(list(AIRFLOW_STATES.values())),
                    )
                )
        elif type(self.attributes["allowed_states"]) == list:
            enum_not_matched = [
                x
                for x in self.attributes["allowed_states"]
                if x not in list(AIRFLOW_STATES.values())
            ]
            if len(enum_not_matched) > 0:
                raise AirflowException(
                    "`%s` is an invalid input of `%s` for `@%s`. Accepted values are %s"
                    % (
                        str(" OR ".join(["'%s'" % i for i in enum_not_matched])),
                        "allowed_states",
                        self.name,
                        ", ".join(list(AIRFLOW_STATES.values())),
                    )
                )
        else:
            self.attributes["allowed_states"] = [AIRFLOW_STATES["SUCCESS"]]

        if self.attributes["execution_delta"] is not None:
            if not isinstance(self.attributes["execution_delta"], timedelta):
                raise AirflowException(
                    "`%s` is an invalid input type of `execution_delta` for `@%s`. Accepted type is `datetime.timedelta`"
                    % (
                        str(type(self.attributes["execution_delta"])),
                        self.name,
                    )
                )
        super().validate()



from metaflow.client import Task
from metaflow import JSONType, namespace
from metaflow.util import resolve_identity
from metaflow.exception import (
    CommandException,
    MetaflowNotFound,
    MetaflowNamespaceMismatch,
)
import webbrowser
import re
from metaflow._vendor import click
import os
import time
import json
import uuid
import signal
import random
from contextlib import contextmanager
from functools import wraps
from metaflow.exception import MetaflowNamespaceMismatch

from .card_datastore import CardDatastore, NUM_SHORT_HASH_CHARS
from .exception import (
    CardClassFoundException,
    IncorrectCardArgsException,
    UnrenderableCardException,
    CardNotPresentException,
    TaskNotFoundException,
)
import traceback
from collections import namedtuple

from .card_resolver import resolve_paths_from_task, resumed_info

id_func = id

CardRenderInfo = namedtuple(
    "CardRenderInfo",
    ["mode", "is_implemented", "data", "timed_out", "timeout_stack_trace"],
)


def open_in_browser(card_path):
    url = "file://" + os.path.abspath(card_path)
    webbrowser.open(url)


def resolve_task_from_pathspec(flow_name, pathspec):
    """
    resolves a task object for the pathspec query on the CLI.
    Args:
        flow_name : (str) : name of flow
        pathspec (str) : can be `stepname` / `runid/stepname` / `runid/stepname/taskid`

    Returns:
        metaflow.Task | None
    """
    from metaflow import Flow, Step, Task
    from metaflow.exception import MetaflowNotFound

    # since pathspec can have many variations.
    pthsplits = pathspec.split("/")
    task = None
    run_id = None
    resolving_from = "task_pathspec"
    if len(pthsplits) == 1:
        # This means stepname
        resolving_from = "stepname"
        latest_run = Flow(flow_name).latest_run
        if latest_run is not None:
            run_id = latest_run.pathspec
            try:
                task = latest_run[pathspec].task
            except KeyError:
                pass
    elif len(pthsplits) == 2:
        # This means runid/stepname
        namespace(None)
        resolving_from = "step_pathspec"
        try:
            task = Step("/".join([flow_name, pathspec])).task
        except MetaflowNotFound:
            pass
    elif len(pthsplits) == 3:
        # this means runid/stepname/taskid
        namespace(None)
        resolving_from = "task_pathspec"
        try:
            task = Task("/".join([flow_name, pathspec]))
        except MetaflowNotFound:
            pass
    else:
        # raise exception for invalid pathspec format
        raise CommandException(
            msg="The PATHSPEC argument should be of the form 'stepname' Or '<runid>/<stepname>' Or '<runid>/<stepname>/<taskid>'"
        )

    if task is None:
        # raise Exception that task could not be resolved for the query.
        raise TaskNotFoundException(pathspec, resolving_from, run_id=run_id)

    return task


def resolve_card(
    ctx,
    pathspec,
    follow_resumed=True,
    hash=None,
    type=None,
    card_id=None,
    no_echo=False,
):
    """Resolves the card path for a query.

    Args:
        ctx: click context object
        pathspec: pathspec can be `stepname` or `runid/stepname` or `runid/stepname/taskid`
        hash (optional): This is to specifically resolve the card via the hash. This is useful when there may be many card with same id or type for a pathspec.
        type : type of card
        card_id : `id` given to card
        no_echo : if set to `True` then supress logs about pathspec resolution.
    Raises:
        CardNotPresentException: No card could be found for the pathspec

    Returns:
        (card_paths, card_datastore, taskpathspec) : Tuple[List[str], CardDatastore, str]
    """
    flow_name = ctx.obj.flow.name
    task = resolve_task_from_pathspec(flow_name, pathspec)
    card_pathspec = task.pathspec
    print_str = "Resolving card: %s" % card_pathspec
    if follow_resumed:
        origin_taskpathspec = resumed_info(task)
        if origin_taskpathspec:
            card_pathspec = origin_taskpathspec
            print_str = "Resolving card resumed from: %s" % origin_taskpathspec

    if not no_echo:
        ctx.obj.echo(print_str, fg="green")
    # to resolve card_id we first check if the identifier is a pathspec and if it is then we check if the `id` is set or not to resolve card_id
    # todo : Fix this with `coalesce function`
    card_paths_found, card_datastore = resolve_paths_from_task(
        ctx.obj.flow_datastore,
        pathspec=card_pathspec,
        type=type,
        hash=hash,
        card_id=card_id,
    )

    if len(card_paths_found) == 0:
        # If there are no files found on the Path then raise an error of
        raise CardNotPresentException(
            card_pathspec, card_hash=hash, card_type=type, card_id=card_id
        )

    return card_paths_found, card_datastore, card_pathspec


@contextmanager
def timeout(time):
    # Register a function to raise a TimeoutError on the signal.
    signal.signal(signal.SIGALRM, raise_timeout)
    # Schedule the signal to be sent after ``time``.
    signal.alarm(time)

    try:
        yield
    finally:
        # Unregister the signal so that it won't be triggered
        # if the timeout is not reached.
        signal.signal(signal.SIGALRM, signal.SIG_IGN)


def raise_timeout(signum, frame):
    raise TimeoutError


def list_available_cards(
    ctx,
    pathspec,
    card_paths,
    card_datastore,
    command="view",
    show_list_as_json=False,
    list_many=False,
    file=None,
):
    # pathspec is full pathspec.
    # todo : create nice response messages on the CLI for cards which were found.
    scriptname = ctx.obj.flow.script_name
    path_tuples = card_datastore.get_card_names(card_paths)
    if show_list_as_json:
        json_arr = [
            dict(id=tup.id, hash=tup.hash, type=tup.type, filename=tup.filename)
            for tup in path_tuples
        ]
        if not list_many:
            # This means that `list_available_cards` is being called once.
            # So we can directly dump the file
            dump_dict = dict(pathspec=pathspec, cards=json_arr)
            if file:
                with open(file, "w") as f:
                    json.dump(dump_dict, f)
            else:
                ctx.obj.echo_always(json.dumps(dump_dict, indent=4), err=False)
        # if you have to list many in json format then return
        return dict(pathspec=pathspec, cards=json_arr)

    if list_many:
        ctx.obj.echo("\tTask: %s" % pathspec.split("/")[-1], fg="green")
    else:
        ctx.obj.echo(
            "Found %d card matching for your query..." % len(path_tuples), fg="green"
        )
    task_pathspec = "/".join(pathspec.split("/")[1:])
    card_list = []
    for path_tuple, file_path in zip(path_tuples, card_paths):
        full_pth = card_datastore.create_full_path(file_path)
        cpr = """
        Card Id: %s
        Card Type: %s
        Card Hash: %s 
        Card Path: %s
        """ % (
            path_tuple.id,
            path_tuple.type,
            path_tuple.hash,
            full_pth,
        )
        card_list.append(cpr)

    random_idx = 0 if len(path_tuples) == 1 else random.randint(0, len(path_tuples) - 1)
    _, randhash, _, file_name = path_tuples[random_idx]
    join_char = "\n\t"
    ctx.obj.echo(join_char.join([""] + card_list) + "\n", fg="blue")

    if command is not None:
        ctx.obj.echo(
            "\n\tExample access from CLI via: \n\t %s\n"
            % make_command(
                scriptname,
                task_pathspec,
                command=command,
                hash=randhash[:NUM_SHORT_HASH_CHARS],
            ),
            fg="yellow",
        )


def make_command(
    script_name,
    taskspec,
    command="get",
    hash=None,
):
    calling_args = ["--hash", hash]
    return " ".join(
        [
            ">>>",
            "python",
            script_name,
            "card",
            command,
            taskspec,
        ]
        + calling_args
    )


def list_many_cards(
    ctx,
    type=None,
    hash=None,
    card_id=None,
    follow_resumed=None,
    as_json=None,
    file=None,
):
    from metaflow import Flow

    flow = Flow(ctx.obj.flow.name)
    run = flow.latest_run
    cards_found = 0
    if not as_json:
        pass
        ctx.obj.echo("Listing cards for run %s" % run.pathspec, fg="green")
    js_list = []
    for step in run:
        step_str_printed = False  # variable to control printing stepname once.
        for task in step:
            try:
                available_card_paths, card_datastore, pathspec = resolve_card(
                    ctx,
                    "/".join(task.pathspec.split("/")[1:]),
                    type=type,
                    hash=hash,
                    card_id=card_id,
                    follow_resumed=follow_resumed,
                    no_echo=True,
                )
                if not step_str_printed and not as_json:
                    ctx.obj.echo("Step : %s" % step.id, fg="green")
                    step_str_printed = True

                js_resp = list_available_cards(
                    ctx,
                    pathspec,
                    available_card_paths,
                    card_datastore,
                    command=None,
                    show_list_as_json=as_json,
                    list_many=True,
                    file=file,
                )
                if as_json:
                    js_list.append(js_resp)
                cards_found += 1
            except CardNotPresentException:
                pass
    if cards_found == 0:
        raise CardNotPresentException(
            run.pathspec, card_hash=hash, card_type=type, card_id=card_id
        )
    if as_json:
        if file:
            with open(file, "w") as f:
                json.dump(js_list, f)
        else:
            ctx.obj.echo_always(json.dumps(js_list, indent=4), err=False)


@click.group()
def cli():
    pass


@cli.group(help="Commands related to @card decorator.")
@click.pass_context
def card(ctx):
    # We set the metadata values here so that top level arguments to --datastore and --metadata
    # Can work with the Metaflow client.
    # If we don't set the metadata here than the metaflow client picks the defaults when calling the `Task`/`Run` objects. These defaults can come from the `config.json` file or based on the `METAFLOW_PROFILE`
    from metaflow import metadata

    setting_metadata = "@".join(
        [ctx.obj.metadata.TYPE, ctx.obj.metadata.default_info()]
    )
    metadata(setting_metadata)
    # set the card root to the datastore according to the configuration.
    root_pth = CardDatastore.get_storage_root(ctx.obj.flow_datastore._storage_impl.TYPE)
    if root_pth is not None:
        ctx.obj.flow_datastore._storage_impl.datastore_root = root_pth


def card_read_options_and_arguments(func):
    @click.option(
        "--hash",
        default=None,
        show_default=True,
        type=str,
        help="Hash of the stored HTML",
    )
    @click.option(
        "--type",
        default=None,
        show_default=True,
        type=str,
        help="Type of card",
    )
    @click.option(
        "--id",
        default=None,
        show_default=True,
        type=str,
        help="Id of the card",
    )
    @click.option(
        "--follow-resumed/--no-follow-resumed",
        default=True,
        show_default=True,
        help="Follow the origin-task-id of resumed tasks to seek cards stored for resumed tasks.",
    )
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)

    return wrapper


def update_card(mf_card, mode, task, data, timeout_value=None):
    """
    This method will be responsible for creating a card/data-update based on the `mode`.
    There are three possible modes taken by this function.
        - render :
            - This will render the "final" card.
            - This mode is passed at task completion.
            - Setting this mode will call the `render` method of a MetaflowCard.
            - It will result in the creation of an HTML page.
        - render_runtime:
            - Setting this mode will render a card during task "runtime".
            - Setting this mode will call the `render_runtime` method of a MetaflowCard.
            - It will result in the creation of an HTML page.
        - refresh:
            - Setting this mode will refresh the data update for a card.
            - We support this mode because rendering a full card can be an expensive operation, but shipping tiny data updates can be cheap.
            - Setting this mode will call the `refresh` method of a MetaflowCard.
            - It will result in the creation of a JSON object.

    Parameters
    ----------
    mf_card : MetaflowCard
        MetaflowCard object which will be used to render the card.
    mode : str
        Mode of rendering the card.
    task : Task
        Task object which will be passed to render the card.
    data : dict
        object created and passed down from `current.card._get_latest_data` method.
        For more information on this object's schema have a look at `current.card._get_latest_data` method.
    timeout_value : int
        Timeout value for rendering the card.

    Returns
    -------
    CardRenderInfo
        - NamedTuple which will contain:
            - `mode`: The mode of rendering the card.
            - `is_implemented`: whether the function was implemented or not.
            - `data` : output from rendering the card (Can be string/dict)
            - `timed_out` : whether the function timed out or not.
            - `timeout_stack_trace` : stack trace of the function if it timed out.
    """

    def _reload_token():
        if "render_seq" not in data:
            return "never"

        if data["render_seq"] == "final":
            # final data update should always trigger a card reload to show
            # the final card, hence a different token for the final update
            return "final"
        elif mf_card.RELOAD_POLICY == mf_card.RELOAD_POLICY_ALWAYS:
            return "render-seq-%s" % data["render_seq"]
        elif mf_card.RELOAD_POLICY == mf_card.RELOAD_POLICY_NEVER:
            return "never"
        elif mf_card.RELOAD_POLICY == mf_card.RELOAD_POLICY_ONCHANGE:
            return mf_card.reload_content_token(task, data)

    def _add_token_html(html):
        if html is None:
            return None
        return html.replace(mf_card.RELOAD_POLICY_TOKEN, _reload_token())

    def _add_token_json(json_msg):
        if json_msg is None:
            return None
        return {
            "reload_token": _reload_token(),
            "data": json_msg,
            "created_on": time.time(),
        }

    def _safe_call_function(func, *args, **kwargs):
        """
        returns (data, is_implemented)
        """
        try:
            return func(*args, **kwargs), True
        except NotImplementedError as e:
            return None, False

    def _call():
        if mode == "render":
            setattr(
                mf_card.__class__,
                "runtime_data",
                property(fget=lambda _, data=data: data),
            )
            output = _add_token_html(mf_card.render(task))
            return CardRenderInfo(
                mode=mode,
                is_implemented=True,
                data=output,
                timed_out=False,
                timeout_stack_trace=None,
            )

        elif mode == "render_runtime":
            # Since many cards created by metaflow users may not have implemented a
            # `render_time` / `refresh` methods, it can result in an exception and thereby
            # creation of error cards (especially for the `render_runtime` method). So instead
            # we will catch the NotImplementedError and return None if users have not implemented it.
            # If there any any other exception from the user code, it should be bubbled to the top level.
            output, is_implemented = _safe_call_function(
                mf_card.render_runtime, task, data
            )
            return CardRenderInfo(
                mode=mode,
                is_implemented=is_implemented,
                data=_add_token_html(output),
                timed_out=False,
                timeout_stack_trace=None,
            )

        elif mode == "refresh":
            output, is_implemented = _safe_call_function(mf_card.refresh, task, data)
            return CardRenderInfo(
                mode=mode,
                is_implemented=is_implemented,
                data=_add_token_json(output),
                timed_out=False,
                timeout_stack_trace=None,
            )

    render_info = None
    if timeout_value is None or timeout_value < 0:
        return _call()
    else:
        try:
            with timeout(timeout_value):
                render_info = _call()
        except TimeoutError:
            stack_trace = traceback.format_exc()
            return CardRenderInfo(
                mode=mode,
                is_implemented=True,
                data=None,
                timed_out=True,
                timeout_stack_trace=stack_trace,
            )
        return render_info


@card.command(help="create a HTML card")
@click.argument("pathspec", type=str)
@click.option(
    "--type",
    default="default",
    show_default=True,
    type=str,
    help="Type of card being created",
)
@click.option(
    "--options",
    default=None,
    show_default=True,
    type=JSONType,
    help="arguments of the card being created.",
)
@click.option(
    "--timeout",
    default=None,
    show_default=True,
    type=int,
    help="Maximum amount of time allowed to create card.",
)
@click.option(
    "--render-error-card",
    default=False,
    is_flag=True,
    help="Upon failing to render a card, render a card holding the stack trace",
)
@click.option(
    "--id",
    default=None,
    show_default=True,
    type=str,
    help="ID of the card",
)
@click.option(
    "--component-file",
    default=None,
    show_default=True,
    type=str,
    help="JSON File with pre-rendered components. (internal)",
)
@click.option(
    "--mode",
    default="render",
    show_default=True,
    type=click.Choice(["render", "render_runtime", "refresh"]),
    help="Rendering mode. (internal)",
)
@click.option(
    "--data-file",
    default=None,
    show_default=True,
    type=str,
    hidden=True,
    help="JSON file containing data to be updated. (internal)",
)
@click.option(
    "--card-uuid",
    default=None,
    show_default=True,
    type=str,
    hidden=True,
    help="Card UUID. (internal)",
)
@click.option(
    "--delete-input-files",
    default=False,
    is_flag=True,
    show_default=True,
    hidden=True,
    help="Delete data-file and component-file after reading. (internal)",
)
@click.pass_context
def create(
    ctx,
    pathspec,
    mode=None,
    type=None,
    options=None,
    timeout=None,
    component_file=None,
    data_file=None,
    render_error_card=False,
    card_uuid=None,
    delete_input_files=None,
    id=None,
):
    card_id = id
    rendered_info = None  # Variable holding all the information which will be rendered
    error_stack_trace = None  # Variable which will keep a track of error

    if len(pathspec.split("/")) != 3:
        raise CommandException(
            msg="Expecting pathspec of form <runid>/<stepname>/<taskid>"
        )
    flowname = ctx.obj.flow.name
    full_pathspec = "/".join([flowname, pathspec])

    graph_dict, _ = ctx.obj.graph.output_steps()

    if card_uuid is None:
        card_uuid = str(uuid.uuid4()).replace("-", "")

    # Components are rendered in a Step and added via `current.card.append` are added here.
    component_arr = []
    if component_file is not None:
        with open(component_file, "r") as f:
            component_arr = json.load(f)
        # Component data used in card runtime is passed in as temporary files which can be deleted after use
        if delete_input_files:
            os.remove(component_file)

    # Load data to be refreshed for runtime cards
    data = {}
    if data_file is not None:
        with open(data_file, "r") as f:
            data = json.load(f)
        # data is passed in as temporary files which can be deleted after use
        if delete_input_files:
            os.remove(data_file)

    task = Task(full_pathspec)
    from metaflow.plugins import CARDS
    from metaflow.plugins.cards.exception import CARD_ID_PATTERN, TYPE_CHECK_REGEX
    from metaflow.cards import ErrorCard

    error_card = ErrorCard
    filtered_cards = [CardClass for CardClass in CARDS if CardClass.type == type]
    card_datastore = CardDatastore(ctx.obj.flow_datastore, pathspec=full_pathspec)

    if len(filtered_cards) == 0 or type is None:
        if render_error_card:
            error_stack_trace = str(CardClassFoundException(type))
        else:
            raise CardClassFoundException(type)

    if len(filtered_cards) > 0:
        filtered_card = filtered_cards[0]
        ctx.obj.echo(
            "Creating new card of type %s with timeout %s"
            % (filtered_card.type, timeout),
            fg="green",
        )
        # If the card is Instantiatable then
        # first instantiate; If instantiation has a TypeError
        # then check for render_error_card and accordingly
        # store the exception as a string or raise the exception
        try:
            if options is not None:
                mf_card = filtered_card(
                    options=options, components=component_arr, graph=graph_dict
                )
            else:
                mf_card = filtered_card(components=component_arr, graph=graph_dict)
        except TypeError as e:
            if render_error_card:
                mf_card = None
                error_stack_trace = str(IncorrectCardArgsException(type, options))
            else:
                raise IncorrectCardArgsException(type, options)

        rendered_content = None
        if mf_card:
            try:
                rendered_info = update_card(
                    mf_card, mode, task, data, timeout_value=timeout
                )
                rendered_content = rendered_info.data
            except:
                rendered_info = CardRenderInfo(
                    mode=mode,
                    is_implemented=True,
                    data=None,
                    timed_out=False,
                    timeout_stack_trace=None,
                )
                if render_error_card:
                    error_stack_trace = str(UnrenderableCardException(type, options))
                else:
                    raise UnrenderableCardException(type, options)

    # In the entire card rendering process, there are a few cases we want to handle:
    # - [mode == "render"]
    #   1. Card is rendered successfully (We store it in the datastore as a HTML file)
    #   2. Card is not rendered successfully and we have --save-error-card flag set to True
    #      (We store it in the datastore as a HTML file with stack trace)
    #   3. Card render timed-out and we have --save-error-card flag set to True
    #      (We store it in the datastore as a HTML file with stack trace)
    #   4. `render` returns nothing and we have --save-error-card flag set to True.
    #       (We store it in the datastore as a HTML file with some message saying you returned nothing)
    # - [mode == "render_runtime"]
    #   1. Card is rendered successfully (We store it in the datastore as a HTML file)
    #   2. `render_runtime` is not implemented but gets called and we have --save-error-card flag set to True.
    #       (We store it in the datastore as a HTML file with some message saying the card should not be a runtime card if this method is not Implemented)
    #   3. `render_runtime` is implemented and raises an exception and we have --save-error-card flag set to True.
    #       (We store it in the datastore as a HTML file with stack trace)
    #   4. `render_runtime` is implemented but returns nothing and we have --save-error-card flag set to True.
    #       (We store it in the datastore as a HTML file with some message saying you returned nothing)
    #   5. `render_runtime` is implemented but times out and we have --save-error-card flag set to True.
    #       (We store it in the datastore as a HTML file with stack trace)
    # - [mode == "refresh"]
    #   1. Data update is created successfully (We store it in the datastore as a JSON file)
    #   2. `refresh` is not implemented. (We do nothing. Don't store anything.)
    #   3. `refresh` is implemented but it raises an exception. (We do nothing. Don't store anything.)
    #   4. `refresh` is implemented but it times out. (We do nothing. Don't store anything.)

    if error_stack_trace is not None and mode != "refresh":
        rendered_content = error_card().render(task, stack_trace=error_stack_trace)
    elif (
        rendered_info.is_implemented
        and rendered_info.timed_out
        and mode != "refresh"
        and render_error_card
    ):
        timeout_stack_trace = (
            "\nCard rendering timed out after %s seconds. "
            "To increase the timeout duration for card rendering, please set the `timeout` parameter in the @card decorator. "
            "\nStack Trace : \n%s"
        ) % (timeout, rendered_info.timeout_stack_trace)
        rendered_content = error_card().render(
            task,
            stack_trace=timeout_stack_trace,
        )
    elif (
        rendered_info.is_implemented
        and rendered_info.data is None
        and render_error_card
        and mode != "refresh"
    ):
        rendered_content = error_card().render(
            task, stack_trace="No information rendered from card of type %s" % type
        )
    elif (
        not rendered_info.is_implemented
        and render_error_card
        and mode == "render_runtime"
    ):
        message = (
            "Card of type %s is a runtime time card with no `render_runtime` implemented. "
            "Please implement `render_runtime` method to allow rendering this card at runtime."
        ) % type
        rendered_content = error_card().render(task, stack_trace=message)

    # todo : should we save native type for error card or error type ?
    if type is not None and re.match(CARD_ID_PATTERN, type) is not None:
        save_type = type
    else:
        save_type = "error"

    # If card_id is doesn't match regex pattern then we will set it as None
    if card_id is not None and re.match(CARD_ID_PATTERN, card_id) is None:
        ctx.obj.echo(
            "`--id=%s` doesn't match REGEX pattern. `--id` will be set to `None`. Please create `--id` of pattern %s."
            % (card_id, TYPE_CHECK_REGEX),
            fg="red",
        )
        card_id = None

    if rendered_content is not None:
        if mode == "refresh":
            card_datastore.save_data(
                card_uuid, save_type, rendered_content, card_id=card_id
            )
            ctx.obj.echo("Data updated", fg="green")
        else:
            card_info = card_datastore.save_card(
                card_uuid, save_type, rendered_content, card_id=card_id
            )
            ctx.obj.echo(
                "Card created with type: %s and hash: %s"
                % (card_info.type, card_info.hash[:NUM_SHORT_HASH_CHARS]),
                fg="green",
            )


@card.command()
@click.argument("pathspec")
@card_read_options_and_arguments
@click.pass_context
def view(
    ctx,
    pathspec,
    hash=None,
    type=None,
    id=None,
    follow_resumed=False,
):
    """
    View the HTML card in browser based on the pathspec.\n
    The pathspec can be of the form:\n
        - <stepname>\n
        - <runid>/<stepname>\n
        - <runid>/<stepname>/<taskid>\n
    """
    card_id = id
    available_card_paths, card_datastore, pathspec = resolve_card(
        ctx,
        pathspec,
        type=type,
        hash=hash,
        card_id=card_id,
        follow_resumed=follow_resumed,
    )
    if len(available_card_paths) == 1:
        open_in_browser(card_datastore.cache_locally(available_card_paths[0]))
    else:
        list_available_cards(
            ctx,
            pathspec,
            available_card_paths,
            card_datastore,
            command="view",
        )


@card.command()
@click.argument("pathspec")
@click.argument("path", required=False)
@card_read_options_and_arguments
@click.pass_context
def get(
    ctx,
    pathspec,
    path,
    hash=None,
    type=None,
    id=None,
    follow_resumed=False,
):
    """
    Get the HTML string of the card based on pathspec.\n
    The pathspec can be of the form:\n
        - <stepname>\n
        - <runid>/<stepname>\n
        - <runid>/<stepname>/<taskid>\n

    Save the card by adding the `path` argument.
    ```
    python myflow.py card get start a.html --type default
    ```
    """
    card_id = id
    available_card_paths, card_datastore, pathspec = resolve_card(
        ctx,
        pathspec,
        type=type,
        hash=hash,
        card_id=card_id,
        follow_resumed=follow_resumed,
    )
    if len(available_card_paths) == 1:
        if path is not None:
            card_datastore.cache_locally(available_card_paths[0], path)
            return
        print(card_datastore.get_card_html(available_card_paths[0]))
    else:
        list_available_cards(
            ctx,
            pathspec,
            available_card_paths,
            card_datastore,
            command="get",
        )


@card.command()
@click.argument("pathspec", required=False)
@card_read_options_and_arguments
@click.option(
    "--as-json",
    default=False,
    is_flag=True,
    help="Print all available cards as a JSON object",
)
@click.option(
    "--file",
    default=None,
    help="Save the available card list to file.",
)
@click.pass_context
def list(
    ctx,
    pathspec=None,
    hash=None,
    type=None,
    id=None,
    follow_resumed=False,
    as_json=False,
    file=None,
):
    card_id = id
    if pathspec is None:
        list_many_cards(
            ctx,
            type=type,
            hash=hash,
            card_id=card_id,
            follow_resumed=follow_resumed,
            as_json=as_json,
            file=file,
        )
        return

    available_card_paths, card_datastore, pathspec = resolve_card(
        ctx,
        pathspec,
        type=type,
        hash=hash,
        card_id=card_id,
        follow_resumed=follow_resumed,
        no_echo=as_json,
    )
    list_available_cards(
        ctx,
        pathspec,
        available_card_paths,
        card_datastore,
        command=None,
        show_list_as_json=as_json,
        file=file,
    )


@card.command(help="Run local card viewer server")
@click.option(
    "--run-id",
    default=None,
    show_default=True,
    type=str,
    help="Run ID of the flow",
)
@click.option(
    "--port",
    default=8324,
    show_default=True,
    type=int,
    help="Port on which Metaflow card viewer server will run",
)
@click.option(
    "--namespace",
    "user_namespace",
    default=None,
    show_default=True,
    type=str,
    help="Namespace of the flow",
)
@click.option(
    "--poll-interval",
    default=5,
    show_default=True,
    type=int,
    help="Polling interval of the card viewer server.",
)
@click.option(
    "--max-cards",
    default=30,
    show_default=True,
    type=int,
    help="Maximum number of cards to be shown at any time by the card viewer server",
)
@click.pass_context
def server(ctx, run_id, port, user_namespace, poll_interval, max_cards):
    from .card_server import create_card_server, CardServerOptions

    user_namespace = resolve_identity() if user_namespace is None else user_namespace
    run, follow_new_runs, _status_message = _get_run_object(
        ctx.obj, run_id, user_namespace
    )
    if _status_message is not None:
        ctx.obj.echo(_status_message, fg="red")
    options = CardServerOptions(
        flow_name=ctx.obj.flow.name,
        run_object=run,
        only_running=False,
        follow_resumed=False,
        flow_datastore=ctx.obj.flow_datastore,
        max_cards=max_cards,
        follow_new_runs=follow_new_runs,
        poll_interval=poll_interval,
    )
    create_card_server(options, port, ctx.obj)


def _get_run_from_cli_set_runid(obj, run_id):
    # This run-id will be set from the command line args.
    # So if we hit a MetaflowNotFound exception / Namespace mismatch then
    # we should raise an exception
    from metaflow import Run

    flow_name = obj.flow.name
    if len(run_id.split("/")) > 1:
        raise CommandException(
            "run_id should NOT be of the form: `<flowname>/<runid>`. Please provide only run-id"
        )
    try:
        pathspec = "%s/%s" % (flow_name, run_id)
        # Since we are looking at all namespaces,
        # we will not
        namespace(None)
        return Run(pathspec)
    except MetaflowNotFound:
        raise CommandException("No run (%s) found for *%s*." % (run_id, flow_name))


def _get_run_object(obj, run_id, user_namespace):
    from metaflow import Flow

    follow_new_runs = True
    flow_name = obj.flow.name

    if run_id is not None:
        follow_new_runs = False
        run = _get_run_from_cli_set_runid(obj, run_id)
        obj.echo("Using run-id %s" % run.pathspec, fg="blue", bold=False)
        return run, follow_new_runs, None

    _msg = "Searching for runs in namespace: %s" % user_namespace
    obj.echo(_msg, fg="blue", bold=False)

    try:
        namespace(user_namespace)
        flow = Flow(pathspec=flow_name)
        run = flow.latest_run
    except MetaflowNotFound:
        # When we have no runs found for the Flow, we need to ensure that
        # if the `follow_new_runs` is set to True; If `follow_new_runs` is set to True then
        # we don't raise the Exception and instead we return None and let the
        # background Thread wait on the Retrieving the run object.
        _status_msg = "No run found for *%s*." % flow_name
        return None, follow_new_runs, _status_msg

    except MetaflowNamespaceMismatch:
        _status_msg = (
            "No run found for *%s* in namespace *%s*. You can switch the namespace using --namespace"
            % (
                flow_name,
                user_namespace,
            )
        )
        return None, follow_new_runs, _status_msg

    obj.echo("Using run-id %s" % run.pathspec, fg="blue", bold=False)
    return run, follow_new_runs, None



import json
import os
import re
import sys
import time
import shutil
import random
import subprocess
from io import RawIOBase, BufferedIOBase
from itertools import chain, starmap
from tempfile import mkdtemp, NamedTemporaryFile
from typing import Dict, Iterable, List, Optional, Tuple, Union, TYPE_CHECKING

from metaflow import FlowSpec
from metaflow.metaflow_current import current
from metaflow.metaflow_config import (
    DATATOOLS_S3ROOT,
    S3_RETRY_COUNT,
    S3_TRANSIENT_RETRY_COUNT,
    S3_SERVER_SIDE_ENCRYPTION,
    TEMPDIR,
)
from metaflow.util import (
    namedtuple_with_defaults,
    is_stringish,
    to_bytes,
    to_unicode,
    to_fileobj,
    url_quote,
    url_unquote,
)
from metaflow.exception import MetaflowException
from metaflow.debug import debug
import metaflow.tracing as tracing

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse

from .s3util import (
    get_s3_client,
    read_in_chunks,
    get_timestamp,
    TRANSIENT_RETRY_START_LINE,
    TRANSIENT_RETRY_LINE_CONTENT,
)

if TYPE_CHECKING:
    import metaflow


def _check_and_init_s3_deps():
    try:
        import boto3
        from boto3.s3.transfer import TransferConfig
    except (ImportError, ModuleNotFoundError):
        raise MetaflowException("You need to install 'boto3' in order to use S3.")


def check_s3_deps(func):
    """The decorated function checks S3 dependencies (as needed for AWS S3 storage backend).
    This includes boto3.
    """

    def _inner_func(*args, **kwargs):
        _check_and_init_s3_deps()
        return func(*args, **kwargs)

    return _inner_func


TEST_INJECT_RETRYABLE_FAILURES = int(
    os.environ.get("METAFLOW_S3_TEST_RETRYABLE_FAILURES", 0)
)


def ensure_unicode(x):
    return None if x is None else to_unicode(x)


PutValue = Union[RawIOBase, BufferedIOBase, str, bytes]

S3GetObject = namedtuple_with_defaults(
    "S3GetObject", [("key", str), ("offset", int), ("length", int)]
)
S3GetObject.__module__ = __name__

S3PutObject = namedtuple_with_defaults(
    "S3PutObject",
    [
        ("key", str),
        ("value", Optional[PutValue]),
        ("path", Optional[str]),
        ("content_type", Optional[str]),
        ("encryption", Optional[str]),
        ("metadata", Optional[Dict[str, str]]),
    ],
    defaults=(None, None, None, None, None),
)
S3PutObject.__module__ = __name__

RangeInfo = namedtuple_with_defaults(
    "RangeInfo",
    [("total_size", int), ("request_offset", int), ("request_length", int)],
    defaults=(0, -1),
)
RangeInfo.__module__ = __name__

RANGE_MATCH = re.compile(r"bytes (?P<start>[0-9]+)-(?P<end>[0-9]+)/(?P<total>[0-9]+)")


class MetaflowS3InvalidObject(MetaflowException):
    headline = "Not a string-like object"


class MetaflowS3URLException(MetaflowException):
    headline = "Invalid address"


class MetaflowS3Exception(MetaflowException):
    headline = "S3 access failed"


class MetaflowS3NotFound(MetaflowException):
    headline = "S3 object not found"


class MetaflowS3AccessDenied(MetaflowException):
    headline = "S3 access denied"


class MetaflowS3InvalidRange(MetaflowException):
    headline = "S3 invalid range"


class S3Object(object):
    """
    This object represents a path or an object in S3,
    with an optional local copy.

    `S3Object`s are not instantiated directly, but they are returned
    by many methods of the `S3` client.
    """

    def __init__(
        self,
        prefix: str,
        url: str,
        path: str,
        size: Optional[int] = None,
        content_type: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
        range_info: Optional[RangeInfo] = None,
        last_modified: Optional[int] = None,
        encryption: Optional[str] = None,
    ):
        # all fields of S3Object should return a unicode object
        prefix, url, path = map(ensure_unicode, (prefix, url, path))

        self._size = size
        self._url = url
        self._path = path
        self._key = None
        self._content_type = content_type
        self._last_modified = last_modified

        self._metadata = None
        if metadata is not None and "metaflow-user-attributes" in metadata:
            self._metadata = json.loads(metadata["metaflow-user-attributes"])

        if range_info and (
            range_info.request_length is None or range_info.request_length < 0
        ):
            self._range_info = RangeInfo(
                range_info.total_size, range_info.request_offset, range_info.total_size
            )
        else:
            self._range_info = range_info

        if path:
            self._size = os.stat(self._path).st_size

        if prefix is None or prefix == url:
            self._key = url
            self._prefix = None
        else:
            self._key = url[len(prefix.rstrip("/")) + 1 :].rstrip("/")
            self._prefix = prefix

        self._encryption = encryption

    @property
    def exists(self) -> bool:
        """
        Does this key correspond to an object in S3?

        Returns
        -------
        bool
            True if this object points at an existing object (file) in S3.
        """
        return self._size is not None

    @property
    def downloaded(self) -> bool:
        """
        Has this object been downloaded?

        If True, the contents can be accessed through `path`, `blob`,
        and `text` properties.

        Returns
        -------
        bool
            True if the contents of this object have been downloaded.
        """
        return bool(self._path)

    @property
    def url(self) -> str:
        """
        S3 location of the object

        Returns
        -------
        str
            The S3 location of this object.
        """
        return self._url

    @property
    def prefix(self) -> str:
        """
        Prefix requested that matches this object.

        Returns
        -------
        str
            Requested prefix
        """
        return self._prefix

    @property
    def key(self) -> str:
        """
        Key corresponds to the key given to the get call that produced
        this object.

        This may be a full S3 URL or a suffix based on what
        was requested.

        Returns
        -------
        str
            Key requested.
        """
        return self._key

    @property
    def path(self) -> Optional[str]:
        """
        Path to a local temporary file corresponding to the object downloaded.

        This file gets deleted automatically when a S3 scope exits.
        Returns None if this S3Object has not been downloaded.

        Returns
        -------
        str
            Local path, if the object has been downloaded.
        """
        return self._path

    @property
    def blob(self) -> Optional[bytes]:
        """
        Contents of the object as a byte string or None if the
        object hasn't been downloaded.

        Returns
        -------
        bytes
            Contents of the object as bytes.
        """
        if self._path:
            with open(self._path, "rb") as f:
                return f.read()

    @property
    def text(self) -> Optional[str]:
        """
        Contents of the object as a string or None if the
        object hasn't been downloaded.

        The object is assumed to contain UTF-8 encoded data.

        Returns
        -------
        str
            Contents of the object as text.
        """
        if self._path:
            return self.blob.decode("utf-8", errors="replace")

    @property
    def size(self) -> Optional[int]:
        """
        Size of the object in bytes.

        Returns None if the key does not correspond to an object in S3.

        Returns
        -------
        int
            Size of the object in bytes, if the object exists.
        """
        return self._size

    @property
    def has_info(self) -> bool:
        """
        Returns true if this `S3Object` contains the content-type MIME header or
        user-defined metadata.

        If False, this means that `content_type`, `metadata`, `range_info` and
        `last_modified` will return None.

        Returns
        -------
        bool
            True if additional metadata is available.
        """
        return (
            self._content_type is not None
            or self._metadata is not None
            or self._range_info is not None
            or self._encryption is not None
        )

    @property
    def metadata(self) -> Optional[Dict[str, str]]:
        """
        Returns a dictionary of user-defined metadata, or None if no metadata
        is defined.

        Returns
        -------
        Dict
            User-defined metadata.
        """
        return self._metadata

    @property
    def content_type(self) -> Optional[str]:
        """
        Returns the content-type of the S3 object or None if it is not defined.

        Returns
        -------
        str
            Content type or None if the content type is undefined.
        """
        return self._content_type

    @property
    def encryption(self) -> Optional[str]:
        """
        Returns the encryption type of the S3 object or None if it is not defined.

        Returns
        -------
        str
            Server-side-encryption type or None if parameter is not set.
        """
        return self._encryption

    @property
    def range_info(self) -> Optional[RangeInfo]:
        """
        If the object corresponds to a partially downloaded object, returns
        information of what was downloaded.

        The returned object has the following fields:
        - `total_size`: Size of the object in S3.
        - `request_offset`: The starting offset.
        - `request_length`: The number of bytes downloaded.

        Returns
        -------
        namedtuple
            An object containing information about the partial download. If
            the `S3Object` doesn't correspond to a partially downloaded file,
            returns None.
        """
        return self._range_info

    @property
    def last_modified(self) -> Optional[int]:
        """
        Returns the last modified unix timestamp of the object.

        Returns
        -------
        int
            Unix timestamp corresponding to the last modified time.
        """
        return self._last_modified

    def __str__(self):
        if self._path:
            return "<S3Object %s (%d bytes, local)>" % (self._url, self._size)
        elif self._size:
            return "<S3Object %s (%d bytes, in S3)>" % (self._url, self._size)
        else:
            return "<S3Object %s (object does not exist)>" % self._url

    def __repr__(self):
        return str(self)


class S3Client(object):
    def __init__(self, s3_role_arn=None, s3_session_vars=None, s3_client_params=None):
        self._s3_client = None
        self._s3_error = None
        self._s3_role = s3_role_arn
        self._s3_session_vars = s3_session_vars
        self._s3_client_params = s3_client_params

    @property
    def client(self):
        if self._s3_client is None:
            self.reset_client()
        return self._s3_client

    @property
    def error(self):
        if self._s3_error is None:
            self.reset_client()
        return self._s3_error

    def reset_client(self):
        self._s3_client, self._s3_error = get_s3_client(
            s3_role_arn=self._s3_role,
            s3_session_vars=self._s3_session_vars,
            s3_client_params=self._s3_client_params,
        )


class S3(object):
    """
    The Metaflow S3 client.

    This object manages the connection to S3 and a temporary diretory that is used
    to download objects. Note that in most cases when the data fits in memory, no local
    disk IO is needed as operations are cached by the operating system, which makes
    operations fast as long as there is enough memory available.

    The easiest way is to use this object as a context manager:
    ```
    with S3() as s3:
        data = [obj.blob for obj in s3.get_many(urls)]
    print(data)
    ```
    The context manager takes care of creating and deleting a temporary directory
    automatically. Without a context manager, you must call `.close()` to delete
    the directory explicitly:
    ```
    s3 = S3()
    data = [obj.blob for obj in s3.get_many(urls)]
    s3.close()
    ```
    You can customize the location of the temporary directory with `tmproot`. It
    defaults to the current working directory.

    To make it easier to deal with object locations, the client can be initialized
    with an S3 path prefix. There are three ways to handle locations:

    1. Use a `metaflow.Run` object or `self`, e.g. `S3(run=self)` which
       initializes the prefix with the global `DATATOOLS_S3ROOT` path, combined
       with the current run ID. This mode makes it easy to version data based
       on the run ID consistently. You can use the `bucket` and `prefix` to
       override parts of `DATATOOLS_S3ROOT`.

    2. Specify an S3 prefix explicitly with `s3root`,
       e.g. `S3(s3root='s3://mybucket/some/path')`.

    3. Specify nothing, i.e. `S3()`, in which case all operations require
       a full S3 url prefixed with `s3://`.

    Parameters
    ----------
    tmproot : str, default: '.'
        Where to store the temporary directory.
    bucket : str, optional
        Override the bucket from `DATATOOLS_S3ROOT` when `run` is specified.
    prefix : str, optional
        Override the path from `DATATOOLS_S3ROOT` when `run` is specified.
    run : FlowSpec or Run, optional
        Derive path prefix from the current or a past run ID, e.g. S3(run=self).
    s3root : str, optional
        If `run` is not specified, use this as the S3 prefix.
    """

    @classmethod
    def get_root_from_config(cls, echo, create_on_absent=True):
        return DATATOOLS_S3ROOT

    @check_s3_deps
    def __init__(
        self,
        tmproot: str = TEMPDIR,
        bucket: Optional[str] = None,
        prefix: Optional[str] = None,
        run: Optional[Union[FlowSpec, "metaflow.Run"]] = None,
        s3root: Optional[str] = None,
        encryption: Optional[str] = S3_SERVER_SIDE_ENCRYPTION,
        **kwargs
    ):
        if run:
            # 1. use a (current) run ID with optional customizations
            if DATATOOLS_S3ROOT is None:
                raise MetaflowS3URLException(
                    "DATATOOLS_S3ROOT is not configured when trying to use S3 storage"
                )
            parsed = urlparse(DATATOOLS_S3ROOT)
            if not bucket:
                bucket = parsed.netloc
            if not prefix:
                prefix = parsed.path
            if isinstance(run, FlowSpec):
                if current.is_running_flow:
                    prefix = os.path.join(prefix, current.flow_name, current.run_id)
                else:
                    raise MetaflowS3URLException(
                        "Initializing S3 with a FlowSpec outside of a running "
                        "flow is not supported."
                    )
            else:
                prefix = os.path.join(prefix, run.parent.id, run.id)

            self._s3root = "s3://%s" % os.path.join(bucket, prefix.strip("/"))
        elif s3root:
            # 2. use an explicit S3 prefix
            parsed = urlparse(to_unicode(s3root))
            if parsed.scheme != "s3":
                raise MetaflowS3URLException(
                    "s3root needs to be an S3 URL prefixed with s3://."
                )
            self._s3root = s3root.rstrip("/")
        else:
            # 3. use the client only with full URLs
            self._s3root = None

        # Note that providing a role, session vars or client params and a client
        # will result in the role/session vars/client params being ignored
        self._s3_role = kwargs.get("role", None)
        self._s3_session_vars = kwargs.get("session_vars", None)
        self._s3_client_params = kwargs.get("client_params", None)
        self._s3_client = kwargs.get(
            "external_client",
            S3Client(
                s3_role_arn=self._s3_role,
                s3_session_vars=self._s3_session_vars,
                s3_client_params=self._s3_client_params,
            ),
        )
        self._s3_inject_failures = kwargs.get(
            "inject_failure_rate", TEST_INJECT_RETRYABLE_FAILURES
        )
        self._tmpdir = mkdtemp(dir=tmproot, prefix="metaflow.s3.")
        self._encryption = encryption

    def __enter__(self) -> "S3":
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        """
        Delete all temporary files downloaded in this context.
        """
        try:
            if not debug.s3client:
                if self._tmpdir:
                    shutil.rmtree(self._tmpdir)
                    self._tmpdir = None
        except:
            pass

    def _url(self, key_value):
        # NOTE: All URLs are handled as Unicode objects (unicode in py2,
        # string in py3) internally. We expect that all URLs passed to this
        # class as either Unicode or UTF-8 encoded byte strings. All URLs
        # returned are Unicode.
        key = getattr(key_value, "key", key_value)
        if self._s3root is None:
            parsed = urlparse(to_unicode(key))
            if parsed.scheme == "s3" and parsed.path:
                return key
            else:
                if current.is_running_flow:
                    raise MetaflowS3URLException(
                        "Specify S3(run=self) when you use S3 inside a running "
                        "flow. Otherwise you have to use S3 with full "
                        "s3:// urls."
                    )
                else:
                    raise MetaflowS3URLException(
                        "Initialize S3 with an 's3root' or 'run' if you don't "
                        "want to specify full s3:// urls."
                    )
        elif key:
            if key.startswith("s3://"):
                raise MetaflowS3URLException(
                    "Don't use absolute S3 URLs when the S3 client is "
                    "initialized with a prefix. URL: %s" % key
                )
            return os.path.join(self._s3root, key)
        else:
            return self._s3root

    def _url_and_range(self, key_value):
        url = self._url(key_value)
        start = getattr(key_value, "offset", None)
        length = getattr(key_value, "length", None)
        range_str = None
        # Range specification are inclusive so getting from offset 500 for 100
        # bytes will read as bytes=500-599
        if start is not None or length is not None:
            if start is None:
                start = 0
            if length is None:
                # Fetch from offset till the end of the file
                range_str = "bytes=%d-" % start
            elif length < 0:
                # Fetch from end; ignore start value here
                range_str = "bytes=-%d" % (-length)
            else:
                # Typical range fetch
                range_str = "bytes=%d-%d" % (start, start + length - 1)
        return url, range_str

    def list_paths(self, keys: Optional[Iterable[str]] = None) -> List[S3Object]:
        """
        List the next level of paths in S3.

        If multiple keys are specified, listings are done in parallel. The returned
        S3Objects have `.exists == False` if the path refers to a prefix, not an
        existing S3 object.

        For instance, if the directory hierarchy is
        ```
        a/0.txt
        a/b/1.txt
        a/c/2.txt
        a/d/e/3.txt
        f/4.txt
        ```
        The `list_paths(['a', 'f'])` call returns
        ```
        a/0.txt (exists == True)
        a/b/ (exists == False)
        a/c/ (exists == False)
        a/d/ (exists == False)
        f/4.txt (exists == True)
        ```

        Parameters
        ----------
        keys : Iterable[str], optional, default None
            List of paths.

        Returns
        -------
        List[S3Object]
            S3Objects under the given paths, including prefixes (directories) that
            do not correspond to leaf objects.
        """

        def _list(keys):
            if keys is None:
                keys = [None]
            urls = ((self._url(key).rstrip("/") + "/", None) for key in keys)
            res = self._read_many_files("list", urls)
            for s3prefix, s3url, size in res:
                if size:
                    yield s3prefix, s3url, None, int(size)
                else:
                    yield s3prefix, s3url, None, None

        return list(starmap(S3Object, _list(keys)))

    def list_recursive(self, keys: Optional[Iterable[str]] = None) -> List[S3Object]:
        """
        List all objects recursively under the given prefixes.

        If multiple keys are specified, listings are done in parallel. All objects
        returned have `.exists == True` as this call always returns leaf objects.

        For instance, if the directory hierarchy is
        ```
        a/0.txt
        a/b/1.txt
        a/c/2.txt
        a/d/e/3.txt
        f/4.txt
        ```
        The `list_paths(['a', 'f'])` call returns
        ```
        a/0.txt (exists == True)
        a/b/1.txt (exists == True)
        a/c/2.txt (exists == True)
        a/d/e/3.txt (exists == True)
        f/4.txt (exists == True)
        ```

        Parameters
        ----------
        keys : Iterable[str], optional, default None
            List of paths.

        Returns
        -------
        List[S3Object]
            S3Objects under the given paths.
        """

        def _list(keys):
            if keys is None:
                keys = [None]
            res = self._read_many_files(
                "list", map(self._url_and_range, keys), recursive=True
            )
            for s3prefix, s3url, size in res:
                yield s3prefix, s3url, None, int(size)

        return list(starmap(S3Object, _list(keys)))

    def info(self, key: Optional[str] = None, return_missing: bool = False) -> S3Object:
        """
        Get metadata about a single object in S3.

        This call makes a single `HEAD` request to S3 which can be
        much faster than downloading all data with `get`.

        Parameters
        ----------
        key : str, optional, default None
            Object to query. It can be an S3 url or a path suffix.
        return_missing : bool, default False
            If set to True, do not raise an exception for a missing key but
            return it as an `S3Object` with `.exists == False`.

        Returns
        -------
        S3Object
            An S3Object corresponding to the object requested. The object
            will have `.downloaded == False`.
        """

        url = self._url(key)
        src = urlparse(url)

        def _info(s3, tmp):
            resp = s3.head_object(Bucket=src.netloc, Key=src.path.lstrip('/"'))
            return {
                "content_type": resp["ContentType"],
                "metadata": resp["Metadata"],
                "size": resp["ContentLength"],
                "last_modified": get_timestamp(resp["LastModified"]),
                "encryption": resp.get("ServerSideEncryption"),
            }

        info_results = None
        try:
            _, info_results = self._one_boto_op(_info, url, create_tmp_file=False)
        except MetaflowS3NotFound:
            if return_missing:
                info_results = None
            else:
                raise
        if info_results:
            return S3Object(
                self._s3root,
                url,
                path=None,
                size=info_results["size"],
                content_type=info_results["content_type"],
                metadata=info_results["metadata"],
                last_modified=info_results["last_modified"],
                encryption=info_results["encryption"],
            )
        return S3Object(self._s3root, url, None)

    def info_many(
        self, keys: Iterable[str], return_missing: bool = False
    ) -> List[S3Object]:
        """
        Get metadata about many objects in S3 in parallel.

        This call makes a single `HEAD` request to S3 which can be
        much faster than downloading all data with `get`.

        Parameters
        ----------
        keys : Iterable[str]
            Objects to query. Each key can be an S3 url or a path suffix.
        return_missing : bool, default False
            If set to True, do not raise an exception for a missing key but
            return it as an `S3Object` with `.exists == False`.

        Returns
        -------
        List[S3Object]
            A list of S3Objects corresponding to the paths requested. The
            objects will have `.downloaded == False`.
        """

        def _head():
            from . import s3op

            res = self._read_many_files(
                "info", map(self._url_and_range, keys), verbose=False, listing=True
            )

            for s3prefix, s3url, fname in res:
                if fname:
                    # We have a metadata file to read from
                    with open(os.path.join(self._tmpdir, fname), "r") as f:
                        info = json.load(f)
                    if info["error"] is not None:
                        # We have an error, we check if it is a missing file
                        if info["error"] == s3op.ERROR_URL_NOT_FOUND:
                            if return_missing:
                                yield self._s3root, s3url, None
                            else:
                                raise MetaflowS3NotFound()
                        elif info["error"] == s3op.ERROR_URL_ACCESS_DENIED:
                            raise MetaflowS3AccessDenied()
                        else:
                            raise MetaflowS3Exception("Got error: %d" % info["error"])
                    else:
                        yield self._s3root, s3url, None, info["size"], info[
                            "content_type"
                        ], info["metadata"], None, info["last_modified"], info[
                            "encryption"
                        ]
                else:
                    # This should not happen; we should always get a response
                    # even if it contains an error inside it
                    raise MetaflowS3Exception("Did not get a response to HEAD")

        return list(starmap(S3Object, _head()))

    def get(
        self,
        key: Optional[Union[str, S3GetObject]] = None,
        return_missing: bool = False,
        return_info: bool = True,
    ) -> S3Object:
        """
        Get a single object from S3.

        Parameters
        ----------
        key : Union[str, S3GetObject], optional, default None
            Object to download. It can be an S3 url, a path suffix, or
            an S3GetObject that defines a range of data to download. If None, or
            not provided, gets the S3 root.
        return_missing : bool, default False
            If set to True, do not raise an exception for a missing key but
            return it as an `S3Object` with `.exists == False`.
        return_info : bool, default True
            If set to True, fetch the content-type and user metadata associated
            with the object at no extra cost, included for symmetry with `get_many`

        Returns
        -------
        S3Object
            An S3Object corresponding to the object requested.
        """
        from boto3.s3.transfer import TransferConfig

        DOWNLOAD_FILE_THRESHOLD = 2 * TransferConfig().multipart_threshold
        DOWNLOAD_MAX_CHUNK = 2 * 1024 * 1024 * 1024 - 1

        url, r = self._url_and_range(key)
        src = urlparse(url)

        def _download(s3, tmp):
            if r:
                resp = s3.get_object(
                    Bucket=src.netloc, Key=src.path.lstrip("/"), Range=r
                )
                # Format is bytes start-end/total; both start and end are inclusive so
                # a 500 bytes file will be `bytes 0-499/500` for the entire file.
                range_result = resp["ContentRange"]
                range_result_match = RANGE_MATCH.match(range_result)
                if range_result_match is None:
                    raise RuntimeError(
                        "Wrong format for ContentRange: %s" % str(range_result)
                    )
                range_result = RangeInfo(
                    int(range_result_match.group("total")),
                    request_offset=int(range_result_match.group("start")),
                    request_length=int(range_result_match.group("end"))
                    - int(range_result_match.group("start"))
                    + 1,
                )
            else:
                resp = s3.get_object(Bucket=src.netloc, Key=src.path.lstrip("/"))
                range_result = None
            sz = resp["ContentLength"]
            if range_result is None:
                range_result = RangeInfo(sz, request_offset=0, request_length=sz)
            if not r and sz > DOWNLOAD_FILE_THRESHOLD:
                # In this case, it is more efficient to use download_file as it
                # will download multiple parts in parallel (it does it after
                # multipart_threshold)
                s3.download_file(src.netloc, src.path.lstrip("/"), tmp)
            else:
                with open(tmp, mode="wb") as t:
                    read_in_chunks(t, resp["Body"], sz, DOWNLOAD_MAX_CHUNK)
            if return_info:
                return {
                    "content_type": resp["ContentType"],
                    # Since Metaflow can also use S3-compatible storage like MinIO,
                    # there maybe some keys missing in the responses given by different S3-compatible object stores.
                    # MinIO is generally accessed via HTTPS, and so it's encrpytion scheme is
                    # TLS/SSL. This is why the `ServerSideEncryption` key is not present
                    # in the response from MinIO.
                    "encryption": resp.get("ServerSideEncryption"),
                    "metadata": resp["Metadata"],
                    "range_result": range_result,
                    "last_modified": get_timestamp(resp["LastModified"]),
                }
            return None

        addl_info = None
        try:
            path, addl_info = self._one_boto_op(_download, url)
        except MetaflowS3NotFound:
            if return_missing:
                path = None
            else:
                raise
        if addl_info:
            return S3Object(
                self._s3root,
                url,
                path,
                content_type=addl_info["content_type"],
                encryption=addl_info["encryption"],
                metadata=addl_info["metadata"],
                range_info=addl_info["range_result"],
                last_modified=addl_info["last_modified"],
            )
        return S3Object(self._s3root, url, path)

    def get_many(
        self,
        keys: Iterable[Union[str, S3GetObject]],
        return_missing: bool = False,
        return_info: bool = True,
    ) -> List[S3Object]:
        """
        Get many objects from S3 in parallel.

        Parameters
        ----------
        keys : Iterable[Union[str, S3GetObject]]
            Objects to download. Each object can be an S3 url, a path suffix, or
            an S3GetObject that defines a range of data to download.
        return_missing : bool, default False
            If set to True, do not raise an exception for a missing key but
            return it as an `S3Object` with `.exists == False`.
        return_info : bool, default True
            If set to True, fetch the content-type and user metadata associated
            with the object at no extra cost, included for symmetry with `get_many`.

        Returns
        -------
        List[S3Object]
            S3Objects corresponding to the objects requested.
        """

        def _get():
            res = self._read_many_files(
                "get",
                map(self._url_and_range, keys),
                allow_missing=return_missing,
                verify=True,
                verbose=False,
                info=return_info,
                listing=True,
            )

            for s3prefix, s3url, fname in res:
                if return_info:
                    if fname:
                        # We have a metadata file to read from
                        with open(
                            os.path.join(self._tmpdir, "%s_meta" % fname), "r"
                        ) as f:
                            info = json.load(f)
                        range_info = info.get("range_result")
                        if range_info:
                            range_info = RangeInfo(
                                range_info["total"],
                                request_offset=range_info["start"],
                                request_length=range_info["end"]
                                - range_info["start"]
                                + 1,
                            )
                            yield self._s3root, s3url, os.path.join(
                                self._tmpdir, fname
                            ), None, info["content_type"], info[
                                "metadata"
                            ], range_info, info[
                                "last_modified"
                            ], info.get(
                                "encryption"
                            )
                    else:
                        yield self._s3root, s3prefix, None
                else:
                    if fname:
                        yield self._s3root, s3url, os.path.join(self._tmpdir, fname)
                    else:
                        # missing entries per return_missing=True
                        yield self._s3root, s3prefix, None

        return list(starmap(S3Object, _get()))

    def get_recursive(
        self, keys: Iterable[str], return_info: bool = False
    ) -> List[S3Object]:
        """
        Get many objects from S3 recursively in parallel.

        Parameters
        ----------
        keys : Iterable[str]
            Prefixes to download recursively. Each prefix can be an S3 url or a path suffix
            which define the root prefix under which all objects are downloaded.
        return_info : bool, default False
            If set to True, fetch the content-type and user metadata associated
            with the object.

        Returns
        -------
        List[S3Object]
            S3Objects stored under the given prefixes.
        """

        def _get():
            res = self._read_many_files(
                "get",
                map(self._url_and_range, keys),
                recursive=True,
                verify=True,
                verbose=False,
                info=return_info,
                listing=True,
            )

            for s3prefix, s3url, fname in res:
                if return_info:
                    # We have a metadata file to read from
                    with open(os.path.join(self._tmpdir, "%s_meta" % fname), "r") as f:
                        info = json.load(f)
                    range_info = info.get("range_result")
                    if range_info:
                        range_info = RangeInfo(
                            range_info["total"],
                            request_offset=range_info["start"],
                            request_length=range_info["end"] - range_info["start"] + 1,
                        )
                    yield self._s3root, s3url, os.path.join(
                        self._tmpdir, fname
                    ), None, info["content_type"], info["metadata"], range_info, info[
                        "last_modified"
                    ], info.get(
                        "encryption"
                    )
                else:
                    yield s3prefix, s3url, os.path.join(self._tmpdir, fname)

        return list(starmap(S3Object, _get()))

    def get_all(self, return_info: bool = False) -> List[S3Object]:
        """
        Get all objects under the prefix set in the `S3` constructor.

        This method requires that the `S3` object is initialized either with `run` or
        `s3root`.

        Parameters
        ----------
        return_info : bool, default False
            If set to True, fetch the content-type and user metadata associated
            with the object.

        Returns
        -------
        Iterable[S3Object]
            S3Objects stored under the main prefix.
        """

        if self._s3root is None:
            raise MetaflowS3URLException(
                "Can't get_all() when S3 is initialized without a prefix"
            )
        else:
            return self.get_recursive([None], return_info)

    def put(
        self,
        key: Union[str, S3PutObject],
        obj: PutValue,
        overwrite: bool = True,
        content_type: Optional[str] = None,
        metadata: Optional[Dict[str, str]] = None,
    ) -> str:
        """
        Upload a single object to S3.

        Parameters
        ----------
        key : Union[str, S3PutObject]
            Object path. It can be an S3 url or a path suffix.
        obj : PutValue
            An object to store in S3. Strings are converted to UTF-8 encoding.
        overwrite : bool, default True
            Overwrite the object if it exists. If set to False, the operation
            succeeds without uploading anything if the key already exists.
        content_type : str, optional, default None
            Optional MIME type for the object.
        metadata : Dict[str, str], optional, default None
            A JSON-encodable dictionary of additional headers to be stored
            as metadata with the object.

        Returns
        -------
        str
            URL of the object stored.
        """

        if isinstance(obj, (RawIOBase, BufferedIOBase)):
            if not obj.readable() or not obj.seekable():
                raise MetaflowS3InvalidObject(
                    "Object corresponding to the key '%s' is not readable or seekable"
                    % key
                )
            blob = obj
        else:
            if not is_stringish(obj):
                raise MetaflowS3InvalidObject(
                    "Object corresponding to the key '%s' is not a string "
                    "or a bytes object." % key
                )
            blob = to_fileobj(obj)
        # We override the close functionality to prevent closing of the
        # file if it is used multiple times when uploading (since upload_fileobj
        # will/may close it on failure)
        real_close = blob.close
        blob.close = lambda: None

        url = self._url(key)
        src = urlparse(url)
        extra_args = None
        if content_type or metadata or self._encryption:
            extra_args = {}
            if content_type:
                extra_args["ContentType"] = content_type
            if metadata:
                extra_args["Metadata"] = {
                    "metaflow-user-attributes": json.dumps(metadata)
                }
            if self._encryption:
                extra_args["ServerSideEncryption"] = self._encryption

        def _upload(s3, _):
            # We make sure we are at the beginning in case we are retrying
            blob.seek(0)

            # We use manual tracing here because boto3 instrumentation
            # has an issue with upload_fileobj losing track of tracing context
            # https://github.com/open-telemetry/opentelemetry-python-contrib/issues/298
            with tracing.traced("s3.upload_fileobj", {"path": src.path}):
                s3.upload_fileobj(
                    blob, src.netloc, src.path.lstrip("/"), ExtraArgs=extra_args
                )

        if overwrite:
            self._one_boto_op(_upload, url, create_tmp_file=False)
            real_close()
            return url
        else:

            def _head(s3, _):
                s3.head_object(Bucket=src.netloc, Key=src.path.lstrip("/"))

            try:
                self._one_boto_op(_head, url, create_tmp_file=False)
            except MetaflowS3NotFound:
                self._one_boto_op(_upload, url, create_tmp_file=False)
            finally:
                real_close()
            return url

    def put_many(
        self,
        key_objs: List[Union[Tuple[str, PutValue], S3PutObject]],
        overwrite: bool = True,
    ) -> List[Tuple[str, str]]:
        """
        Upload many objects to S3.

        Each object to be uploaded can be specified in two ways:

        1. As a `(key, obj)` tuple where `key` is a string specifying
           the path and `obj` is a string or a bytes object.

        2. As a `S3PutObject` which contains additional metadata to be
           stored with the object.

        Parameters
        ----------
        key_objs : List[Union[Tuple[str, PutValue], S3PutObject]]
            List of key-object pairs to upload.
        overwrite : bool, default True
            Overwrite the object if it exists. If set to False, the operation
            succeeds without uploading anything if the key already exists.

        Returns
        -------
        List[Tuple[str, str]]
            List of `(key, url)` pairs corresponding to the objects uploaded.
        """

        def _store():
            for key_obj in key_objs:
                if isinstance(key_obj, S3PutObject):
                    key = key_obj.key
                    obj = key_obj.value
                else:
                    key = key_obj[0]
                    obj = key_obj[1]
                store_info = {
                    "key": key,
                    "content_type": getattr(key_obj, "content_type", None),
                }
                metadata = getattr(key_obj, "metadata", None)
                if metadata:
                    store_info["metadata"] = {
                        "metaflow-user-attributes": json.dumps(metadata)
                    }
                if self._encryption:
                    store_info["encryption"] = self._encryption
                if isinstance(obj, (RawIOBase, BufferedIOBase)):
                    if not obj.readable() or not obj.seekable():
                        raise MetaflowS3InvalidObject(
                            "Object corresponding to the key '%s' is not readable or seekable"
                            % key
                        )
                else:
                    if not is_stringish(obj):
                        raise MetaflowS3InvalidObject(
                            "Object corresponding to the key '%s' is not a string "
                            "or a bytes object." % key
                        )
                    obj = to_fileobj(obj)
                with NamedTemporaryFile(
                    dir=self._tmpdir,
                    delete=False,
                    mode="wb",
                    prefix="metaflow.s3.put_many.",
                ) as tmp:
                    tmp.write(obj.read())
                    tmp.close()
                    yield tmp.name, self._url(key), store_info

        return self._put_many_files(_store(), overwrite)

    def put_files(
        self,
        key_paths: List[Union[Tuple[str, PutValue], S3PutObject]],
        overwrite: bool = True,
    ) -> List[Tuple[str, str]]:
        """
        Upload many local files to S3.

        Each file to be uploaded can be specified in two ways:

        1. As a `(key, path)` tuple where `key` is a string specifying
           the S3 path and `path` is the path to a local file.

        2. As a `S3PutObject` which contains additional metadata to be
           stored with the file.

        Parameters
        ----------
        key_paths :  List[Union[Tuple[str, PutValue], S3PutObject]]
            List of files to upload.
        overwrite : bool, default True
            Overwrite the object if it exists. If set to False, the operation
            succeeds without uploading anything if the key already exists.

        Returns
        -------
        List[Tuple[str, str]]
            List of `(key, url)` pairs corresponding to the files uploaded.
        """

        def _check():
            for key_path in key_paths:
                if isinstance(key_path, S3PutObject):
                    key = key_path.key
                    path = key_path.path
                else:
                    key = key_path[0]
                    path = key_path[1]
                store_info = {
                    "key": key,
                    "content_type": getattr(key_path, "content_type", None),
                }
                metadata = getattr(key_path, "metadata", None)
                if metadata:
                    store_info["metadata"] = {
                        "metaflow-user-attributes": json.dumps(metadata)
                    }
                if self._encryption:
                    store_info["encryption"] = self._encryption
                if not os.path.exists(path):
                    raise MetaflowS3NotFound("Local file not found: %s" % path)
                yield path, self._url(key), store_info

        return self._put_many_files(_check(), overwrite)

    def _one_boto_op(self, op, url, create_tmp_file=True):
        error = ""
        for i in range(S3_RETRY_COUNT + 1):
            tmp = None
            if create_tmp_file:
                tmp = NamedTemporaryFile(
                    dir=self._tmpdir, prefix="metaflow.s3.one_file.", delete=False
                )
            try:
                side_results = op(self._s3_client.client, tmp.name if tmp else None)
                return tmp.name if tmp else None, side_results
            except self._s3_client.error as err:
                from . import s3op

                error_code = s3op.normalize_client_error(err)
                if error_code == 404:
                    raise MetaflowS3NotFound(url)
                elif error_code == 403:
                    raise MetaflowS3AccessDenied(url)
                elif error_code == 416:
                    raise MetaflowS3InvalidRange(err)
                elif error_code == "NoSuchBucket":
                    raise MetaflowS3URLException("Specified S3 bucket doesn't exist.")
                error = str(err)
            except Exception as ex:
                # TODO specific error message for out of disk space
                error = str(ex)
            if tmp:
                os.unlink(tmp.name)
            self._s3_client.reset_client()
            # only sleep if retries > 0
            if S3_RETRY_COUNT > 0:
                self._jitter_sleep(i)
        raise MetaflowS3Exception(
            "S3 operation failed.\n" "Key requested: %s\n" "Error: %s" % (url, error)
        )

    # add some jitter to make sure retries are not synchronized
    def _jitter_sleep(self, trynum, multiplier=2):
        interval = multiplier**trynum + random.randint(0, 10)
        time.sleep(interval)

    # NOTE: re: _read_many_files and _put_many_files
    # All file IO is through binary files - we write bytes, we read
    # bytes. All inputs and outputs from these functions are Unicode.
    # Conversion between bytes and unicode is done through
    # and url_unquote.
    def _read_many_files(self, op, prefixes_and_ranges, **options):
        prefixes_and_ranges = list(prefixes_and_ranges)
        with NamedTemporaryFile(
            dir=self._tmpdir,
            mode="wb",
            delete=not debug.s3client,
            prefix="metaflow.s3.inputs.",
        ) as inputfile:
            inputfile.write(
                b"\n".join(
                    [
                        b" ".join([url_quote(prefix)] + ([url_quote(r)] if r else []))
                        for prefix, r in prefixes_and_ranges
                    ]
                )
            )
            inputfile.flush()
            stdout_lines, stderr = self._s3op_with_retries(
                op, inputs=inputfile.name, **options
            )
            if stderr:
                raise MetaflowS3Exception(
                    "Getting S3 files failed.\n"
                    "First prefix requested: %s\n"
                    "Error: %s" % (prefixes_and_ranges[0], stderr)
                )
            else:
                for line in stdout_lines:
                    yield tuple(map(url_unquote, line.strip(b"\n").split(b" ")))

    def _put_many_files(self, url_info, overwrite):
        url_info = list(url_info)
        url_dicts = [
            dict(
                chain([("local", os.path.realpath(local)), ("url", url)], info.items())
            )
            for local, url, info in url_info
        ]

        with NamedTemporaryFile(
            dir=self._tmpdir,
            mode="wb",
            delete=not debug.s3client,
            prefix="metaflow.s3.put_inputs.",
        ) as inputfile:
            lines = [to_bytes(json.dumps(x)) for x in url_dicts]
            inputfile.write(b"\n".join(lines))
            inputfile.flush()
            stdout_lines, stderr = self._s3op_with_retries(
                "put",
                inputs=inputfile.name,
                verbose=False,
                overwrite=overwrite,
                listing=True,
            )
            if stderr:
                raise MetaflowS3Exception(
                    "Uploading S3 files failed.\n"
                    "First key: %s\n"
                    "Error: %s" % (url_info[0][2]["key"], stderr)
                )
            else:
                urls = set()
                for line in stdout_lines:
                    url, _, _ = map(url_unquote, line.strip(b"\n").split(b" "))
                    urls.add(url)
                return [(info["key"], url) for _, url, info in url_info if url in urls]

    def _s3op_with_retries(self, mode, **options):
        from . import s3op

        # High level note on what this function does:
        #  - perform s3op (which calls s3op.py in a subprocess to parallelize the
        #    operation). Typically this operation has several inputs (for example,
        #    multiple files to get or put)
        #  - the result of this operation can be either:
        #    - a known permanent failure (access denied for example) in which case we
        #      return this failure.
        #    - a known transient failure (SlowDown for example) in which case we will
        #      retry *only* the inputs that have this transient failure.
        #    - an unknown failure (something went wrong but we cannot say if it was
        #      a known permanent failure or something else). In this case, we retry
        #      the operation completely.
        #
        # There are therefore two retry counts:
        #  - the transient failure retry count: how many times do we try on known
        #    transient errors
        #  - the top-level retry count: how many times do we try on unknown failures
        #
        # Note that, if the operation runs out of transient failure retries, it will
        # count as an "unknown" failure (ie: it will be retried according to the
        # outer top-level retry count). In other words, you can potentially have
        # transient_retry_count * retry_count tries).
        # Finally, if on transient failures, we make NO progress (ie: no input is
        # successfully processed), that counts as an "unknown" failure.
        cmdline = [sys.executable, os.path.abspath(s3op.__file__), mode]
        recursive_get = False
        for key, value in options.items():
            key = key.replace("_", "-")
            if isinstance(value, bool):
                if value:
                    if mode == "get" and key == "recursive":
                        # We make a note of this because for transient retries, we
                        # don't pass the recursive flag since we already did all the
                        # listing we needed
                        recursive_get = True
                    else:
                        cmdline.append("--%s" % key)
                else:
                    cmdline.append("--no-%s" % key)
            elif key == "inputs":
                base_input_filename = value
            else:
                cmdline.extend(("--%s" % key, value))
        if self._s3_role is not None:
            cmdline.extend(("--s3role", self._s3_role))
        if self._s3_session_vars is not None:
            cmdline.extend(("--s3sessionvars", json.dumps(self._s3_session_vars)))
        if self._s3_client_params is not None:
            cmdline.extend(("--s3clientparams", json.dumps(self._s3_client_params)))

        def _inject_failure_rate():
            # list mode does not do retries on transient failures (there is no
            # SlowDown handling) so we never inject a failure rate
            if mode == "list":
                return 0
            # Otherwise, we cap the failure rate at 90%
            return min(90, self._s3_inject_failures)

        retry_count = 0  # Number of retries (excluding transient failures)
        transient_retry_count = 0  # Number of transient retries (per top-level retry)
        inject_failures = _inject_failure_rate()
        out_lines = []  # List to contain the lines returned by _s3op_with_retries
        pending_retries = (
            []
        )  # Inputs that need to be retried due to a transient failure
        loop_count = 0
        last_ok_count = 0  # Number of inputs that were successful in the last try
        total_ok_count = 0  # Total number of OK inputs

        def _reset():
            nonlocal transient_retry_count, inject_failures, out_lines, pending_retries
            nonlocal loop_count, last_ok_count, total_ok_count
            transient_retry_count = 0
            inject_failures = _inject_failure_rate()
            if mode != "put":
                # For put, even after retries, we keep around whatever we already
                # uploaded. This is because uploading with overwrite=False is not
                # an idempotent operation and so some files could be uploaded during
                # the first try which we should report back.
                out_lines = []
            pending_retries = []
            loop_count = 0
            last_ok_count = 0
            total_ok_count = 0  # Reset to zero even if we keep out_lines

        def _update_out_lines(out_lines, ok_lines, resize=False):
            if resize:
                # This is the first time around; we make the list big enough. Typically,
                # there is nothing in out_lines but in some cases (a retry after a
                # partial result), there may be stuff in it
                out_lines.extend([None] * (len(ok_lines) - len(out_lines)))
            for l in ok_lines:
                idx, rest = l.split(b" ", maxsplit=1)
                if rest.decode(encoding="utf-8") != TRANSIENT_RETRY_LINE_CONTENT:
                    # Update the proper location in the out_lines array; we maintain
                    # position as if transient retries did not exist. This
                    # makes sure that order is respected even in the presence of
                    # transient retries.
                    out_lines[int(idx.decode(encoding="utf-8"))] = rest

        def try_s3_op(last_ok_count, pending_retries, out_lines, inject_failures):
            # NOTE: Make sure to update pending_retries and out_lines in place
            addl_cmdline = []
            if len(pending_retries) == 0 and recursive_get:
                # First time around (or after a fatal failure)
                addl_cmdline = ["--recursive"]
            with NamedTemporaryFile(
                dir=self._tmpdir,
                mode="wb+",
                delete=not debug.s3client,
                prefix="metaflow.s3op.stderr.",
            ) as stderr:
                with NamedTemporaryFile(
                    dir=self._tmpdir,
                    mode="wb",
                    delete=not debug.s3client,
                    prefix="metaflow.s3op.transientretry.",
                ) as tmp_input:
                    if len(pending_retries) > 0:
                        # We try a little bit more than the previous success (to still
                        # be aggressive but not too much). If there is a lot of
                        # transient errors and we are having issues pushing through
                        # things, this will shrink more and more until we are doing a
                        # single operation at a time. If things start going better, it
                        # will increase by 20% every round.
                        max_count = min(int(last_ok_count * 1.2), len(pending_retries))
                        tmp_input.writelines(pending_retries[:max_count])
                        tmp_input.flush()
                        debug.s3client_exec(
                            "Have %d pending; succeeded in %d => trying for %d and "
                            "leaving %d for the next round"
                            % (
                                len(pending_retries),
                                last_ok_count,
                                max_count,
                                len(pending_retries) - max_count,
                            )
                        )
                        del pending_retries[:max_count]

                        input_filename = tmp_input.name
                    else:
                        input_filename = base_input_filename

                    addl_cmdline.extend(["--inputs", input_filename])

                    # Check if we want to inject failures (for testing)
                    if inject_failures > 0:
                        addl_cmdline.extend(["--inject-failure", str(inject_failures)])
                        # Logic here is to have higher and lower failure rates to try to
                        # exercise as much of the code as possible. The failure rate
                        # trends towards 0.
                        if loop_count % 2 == 0:
                            inject_failures = int(inject_failures / 3)
                        else:
                            # We cap at 90 (and not 100) for injection of failures to
                            # reduce the likelihood of having flaky test. If the
                            # failure injection rate is too high, this can cause actual
                            # retries more often and then lead to too many actual
                            # retries
                            inject_failures = min(90, int(inject_failures * 1.5))
                    try:
                        debug.s3client_exec(cmdline + addl_cmdline)
                        # Run the operation.
                        env = os.environ.copy()
                        tracing.inject_tracing_vars(env)
                        env["METAFLOW_ESCAPE_HATCH_WARNING"] = "False"
                        stdout = subprocess.check_output(
                            cmdline + addl_cmdline,
                            cwd=self._tmpdir,
                            stderr=stderr.file,
                            env=env,
                        )
                        # Here we did not have any error -- transient or otherwise.
                        ok_lines = stdout.splitlines()
                        _update_out_lines(out_lines, ok_lines, resize=loop_count == 0)
                        return (len(ok_lines), 0, inject_failures, None)
                    except subprocess.CalledProcessError as ex:
                        if ex.returncode == s3op.ERROR_TRANSIENT:
                            # In this special case, we failed transiently on *some* of
                            # the files but not necessarily all. This is typically
                            # caused by limits on the number of operations that can
                            # occur per second or some other temporary limitation.
                            # We will retry only those that we failed on and we will not
                            # count this as a retry *unless* we are making no forward
                            # progress. In effect, we consider that as long as *some*
                            # operations are going through, we should just keep going as
                            # if it was a single operation.
                            ok_lines = ex.stdout.splitlines()
                            stderr.seek(0)
                            do_output = False
                            retry_lines = []
                            for l in stderr:
                                if do_output:
                                    retry_lines.append(l)
                                    continue
                                if (
                                    l.decode(encoding="utf-8")
                                    == "%s\n" % TRANSIENT_RETRY_START_LINE
                                ):
                                    # Look for a special marker as the start of the
                                    # "failed inputs that need to be retried"
                                    do_output = True
                            stderr.seek(0)
                            if do_output is False:
                                return (
                                    0,
                                    0,
                                    inject_failures,
                                    "Could not find inputs to retry",
                                )
                            else:
                                _update_out_lines(
                                    out_lines, ok_lines, resize=loop_count == 0
                                )
                                pending_retries.extend(retry_lines)

                                return (
                                    len(ok_lines),
                                    len(retry_lines),
                                    inject_failures,
                                    None,
                                )

                        # Here, this is a "normal" failure that we need to send back up.
                        # These failures are not retried.
                        stderr.seek(0)
                        err_out = stderr.read().decode("utf-8", errors="replace")
                        stderr.seek(0)
                        if ex.returncode == s3op.ERROR_URL_NOT_FOUND:
                            raise MetaflowS3NotFound(err_out)
                        elif ex.returncode == s3op.ERROR_URL_ACCESS_DENIED:
                            raise MetaflowS3AccessDenied(err_out)
                        elif ex.returncode == s3op.ERROR_INVALID_RANGE:
                            raise MetaflowS3InvalidRange(err_out)

                        # Here, this is some other error that we will retry. We still
                        # update the successful lines
                        ok_lines = ex.stdout.splitlines()
                        _update_out_lines(out_lines, ok_lines, resize=loop_count == 0)
                        return 0, 0, inject_failures, err_out

        while retry_count <= S3_RETRY_COUNT:
            (
                last_ok_count,
                last_retry_count,
                inject_failures,
                err_out,
            ) = try_s3_op(last_ok_count, pending_retries, out_lines, inject_failures)
            if err_out or (
                last_retry_count != 0
                and (
                    last_ok_count == 0
                    or transient_retry_count > S3_TRANSIENT_RETRY_COUNT
                )
            ):
                # We had a fatal failure (err_out is not None)
                # or we made no progress (last_ok_count is 0)
                # or we are out of transient retries
                # so we will restart from scratch (being very conservative)
                retry_count += 1
                err_msg = err_out
                if err_msg is None and last_ok_count == 0:
                    err_msg = "No progress"
                if err_msg is None:
                    err_msg = "Too many transient errors"
                print(
                    "S3 non-transient error (attempt #%d): %s" % (retry_count, err_msg)
                )
                _reset()
                if retry_count <= S3_RETRY_COUNT:
                    self._jitter_sleep(retry_count)
                continue
            elif last_retry_count != 0:
                # During our last try, we did not manage to process everything we wanted
                # due to a transient failure so we try again.
                transient_retry_count += 1
                total_ok_count += last_ok_count
                print(
                    "Transient S3 failure (attempt #%d) -- total success: %d, "
                    "last attempt %d/%d -- remaining: %d"
                    % (
                        transient_retry_count,
                        total_ok_count,
                        last_ok_count,
                        last_ok_count + last_retry_count,
                        len(pending_retries),
                    )
                )
                if inject_failures == 0:
                    # Don't sleep when we are "faking" the failures
                    self._jitter_sleep(transient_retry_count)

            loop_count += 1
            # If we have no more things to try, we break out of the loop.
            if len(pending_retries) == 0:
                break

        # At this point, we check out_lines; strip None which can happen for puts that
        # didn't upload files
        return [o for o in out_lines if o is not None], err_out



import os
import sys
import time

from metaflow.exception import MetaflowException

from .kubernetes_job import KubernetesJob


CLIENT_REFRESH_INTERVAL_SECONDS = 300


class KubernetesClientException(MetaflowException):
    headline = "Kubernetes client error"


class KubernetesClient(object):
    def __init__(self):
        try:
            # Kubernetes is a soft dependency.
            from kubernetes import client, config
        except (NameError, ImportError):
            raise KubernetesClientException(
                "Could not import module 'kubernetes'.\n\nInstall Kubernetes "
                "Python package (https://pypi.org/project/kubernetes/) first.\n"
                "You can install the module by executing - "
                "%s -m pip install kubernetes\n"
                "or equivalent through your favorite Python package manager."
                % sys.executable
            )
        self._refresh_client()

    def _refresh_client(self):
        from kubernetes import client, config

        if os.getenv("KUBECONFIG"):
            # There are cases where we're running inside a pod, but can't use
            # the kubernetes client for that pod's cluster: for example when
            # running in Bitbucket Cloud or other CI system.
            # In this scenario, the user can set a KUBECONFIG environment variable
            # to load the kubeconfig, regardless of whether we're in a pod or not.
            config.load_kube_config()
        elif os.getenv("KUBERNETES_SERVICE_HOST"):
            # We are inside a pod, authenticate via ServiceAccount assigned to us
            config.load_incluster_config()
        else:
            # Default to using kubeconfig, likely $HOME/.kube/config
            # TODO (savin):
            #  1. Support generating kubeconfig on the fly using boto3
            #  2. Support auth via OIDC - https://docs.aws.amazon.com/eks/latest/userguide/authenticate-oidc-identity-provider.html
            config.load_kube_config()
        self._client = client
        self._client_refresh_timestamp = time.time()

    def get(self):
        if (
            time.time() - self._client_refresh_timestamp
            > CLIENT_REFRESH_INTERVAL_SECONDS
        ):
            self._refresh_client()

        return self._client

    def job(self, **kwargs):
        return KubernetesJob(self, **kwargs)



import collections
import glob
import json
import os
import random
import tempfile
import time
from collections import namedtuple

from metaflow.exception import MetaflowInternalError, MetaflowTaggingError
from metaflow.metadata.metadata import ObjectOrder
from metaflow.metaflow_config import DATASTORE_LOCAL_DIR
from metaflow.metadata import MetadataProvider
from metaflow.tagging_util import MAX_USER_TAG_SET_SIZE, validate_tags


class LocalMetadataProvider(MetadataProvider):
    TYPE = "local"

    def __init__(self, environment, flow, event_logger, monitor):
        super(LocalMetadataProvider, self).__init__(
            environment, flow, event_logger, monitor
        )

    @classmethod
    def compute_info(cls, val):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        v = os.path.realpath(os.path.join(val, DATASTORE_LOCAL_DIR))
        if os.path.isdir(v):
            LocalStorage.datastore_root = v
            return val
        raise ValueError(
            "Could not find directory %s in directory %s" % (DATASTORE_LOCAL_DIR, val)
        )

    @classmethod
    def default_info(cls):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        def print_clean(line, **kwargs):
            print(line)

        v = LocalStorage.get_datastore_root_from_config(
            print_clean, create_on_absent=False
        )
        if v is None:
            return (
                "<No %s directory found in current working tree>" % DATASTORE_LOCAL_DIR
            )
        return os.path.dirname(v)

    def version(self):
        return "local"

    def new_run_id(self, tags=None, sys_tags=None):
        # We currently just use the timestamp to create an ID. We can be reasonably certain
        # that it is unique and this makes it possible to do without coordination or
        # reliance on POSIX locks in the filesystem.
        run_id = "%d" % (time.time() * 1e6)
        self._new_run(run_id, tags, sys_tags)
        return run_id

    def register_run_id(self, run_id, tags=None, sys_tags=None):
        try:
            # This metadata provider only generates integer IDs so if this is
            # an integer, we don't register it again (since it was "registered"
            # on creation). However, some IDs are created outside the metadata
            # provider and need to be properly registered
            int(run_id)
            return False
        except ValueError:
            return self._new_run(run_id, tags, sys_tags)

    def new_task_id(self, run_id, step_name, tags=None, sys_tags=None):
        self._task_id_seq += 1
        task_id = str(self._task_id_seq)
        self._new_task(run_id, step_name, task_id, tags=tags, sys_tags=sys_tags)
        return task_id

    def register_task_id(
        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
    ):
        try:
            # Same logic as register_run_id
            int(task_id)
        except ValueError:
            return self._new_task(
                run_id,
                step_name,
                task_id,
                attempt=attempt,
                tags=tags,
                sys_tags=sys_tags,
            )
        else:
            self._register_system_metadata(run_id, step_name, task_id, attempt)
            return False

    def register_data_artifacts(
        self, run_id, step_name, task_id, attempt_id, artifacts
    ):
        meta_dir = self._create_and_get_metadir(
            self._flow_name, run_id, step_name, task_id
        )
        artlist = self._artifacts_to_json(
            run_id, step_name, task_id, attempt_id, artifacts
        )
        artdict = {"%d_artifact_%s" % (attempt_id, art["name"]): art for art in artlist}
        self._save_meta(meta_dir, artdict)

    def register_metadata(self, run_id, step_name, task_id, metadata):
        meta_dir = self._create_and_get_metadir(
            self._flow_name, run_id, step_name, task_id
        )
        metalist = self._metadata_to_json(run_id, step_name, task_id, metadata)
        ts = int(round(time.time() * 1000))
        metadict = {
            "sysmeta_%s_%d" % (meta["field_name"], ts): meta for meta in metalist
        }
        self._save_meta(meta_dir, metadict)

    @classmethod
    def _mutate_user_tags_for_run(
        cls, flow_id, run_id, tags_to_add=None, tags_to_remove=None
    ):
        MutationResult = namedtuple(
            "MutationResult", field_names="tags_are_consistent tags"
        )

        def _optimistically_mutate():
            # get existing tags
            run = LocalMetadataProvider.get_object(
                "run", "self", {}, None, flow_id, run_id
            )
            if not run:
                raise MetaflowTaggingError(
                    msg="Run not found (%s, %s)" % (flow_id, run_id)
                )
            existing_user_tag_set = frozenset(run["tags"])
            existing_system_tag_set = frozenset(run["system_tags"])
            tags_to_remove_set = frozenset(tags_to_remove)
            # make sure no existing system tags get added as a user tag
            tags_to_add_set = frozenset(tags_to_add) - existing_system_tag_set

            # from this point on we work with sets of tags only

            if tags_to_remove_set & existing_system_tag_set:
                raise MetaflowTaggingError(
                    msg="Cannot remove a tag that is an existing system tag (%s)"
                    % str(sorted(tags_to_remove_set & existing_system_tag_set))
                )

            # remove tags first, then add
            next_user_tags_set = (
                existing_user_tag_set - tags_to_remove_set
            ) | tags_to_add_set
            # we think it will be a no-op, so let's return right away
            if next_user_tags_set == existing_user_tag_set:
                return MutationResult(
                    tags=next_user_tags_set,
                    tags_are_consistent=True,
                )

            validate_tags(next_user_tags_set, existing_tags=existing_user_tag_set)

            # write new tag set to file system
            LocalMetadataProvider._persist_tags_for_run(
                flow_id, run_id, next_user_tags_set, existing_system_tag_set
            )

            # read tags back from file system to see if our optimism is misplaced
            # I.e. did a concurrent mutate overwrite our change
            run = LocalMetadataProvider.get_object(
                "run", "self", {}, None, flow_id, run_id
            )
            if not run:
                raise MetaflowTaggingError(
                    msg="Run not found for read-back check (%s, %s)" % (flow_id, run_id)
                )
            final_tag_set = frozenset(run["tags"])
            if tags_to_add_set - final_tag_set:
                return MutationResult(tags=final_tag_set, tags_are_consistent=False)
            if (
                tags_to_remove_set & final_tag_set
            ) - tags_to_add_set:  # Remove before add, remember?  Account for this
                return MutationResult(tags=final_tag_set, tags_are_consistent=False)

            return MutationResult(tags=final_tag_set, tags_are_consistent=True)

        tries = 1
        # try up to 5 times, with a gentle exponential backoff (1.1-1.3x)
        while True:
            mutation_result = _optimistically_mutate()
            if mutation_result.tags_are_consistent:
                return mutation_result.tags
            if tries >= 5:
                break
            time.sleep(0.3 * random.uniform(1.1, 1.3) ** tries)
            tries += 1
        raise MetaflowTaggingError(
            "Tagging failed due to too many conflicting updates from other processes"
        )

    @classmethod
    def _get_object_internal(
        cls, obj_type, obj_order, sub_type, sub_order, filters, attempt, *args
    ):
        # This is guaranteed by MetaflowProvider.get_object(), sole intended caller
        if obj_type in ("metadata", "self"):
            raise MetaflowInternalError(msg="Type %s is not allowed" % obj_type)

        if obj_type not in ("root", "flow", "run", "step", "task", "artifact"):
            raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)

        from metaflow.plugins.datastores.local_storage import LocalStorage

        if obj_type == "artifact":
            # Artifacts are actually part of the tasks in the filesystem
            # E.g. we get here for (obj_type, sub_type) == (artifact, self)
            obj_type = "task"
            sub_type = "artifact"
            sub_order = obj_order
            obj_order = obj_order - 1

        if obj_type != ObjectOrder.order_to_type(obj_order):
            raise MetaflowInternalError(
                "Object type order mismatch %s %s"
                % (obj_type, ObjectOrder.order_to_type(obj_order))
            )
        if sub_type != ObjectOrder.order_to_type(sub_order):
            raise MetaflowInternalError(
                "Sub type order mismatch %s %s"
                % (sub_type, ObjectOrder.order_to_type(sub_order))
            )

        RUN_ORDER = ObjectOrder.type_to_order("run")

        if obj_type not in ("root", "flow", "run", "step", "task"):
            raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)

        # Special handling of self, artifact, and metadata
        if sub_type == "self":
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            if meta_path is None:
                return None
            self_file = os.path.join(meta_path, "_self.json")
            if os.path.isfile(self_file):
                obj = MetadataProvider._apply_filter(
                    [LocalMetadataProvider._read_json_file(self_file)], filters
                )[0]
                # For non-descendants of a run, we are done

                if obj_order <= RUN_ORDER:
                    return obj

                if obj_type not in ("step", "task"):
                    raise MetaflowInternalError(
                        msg="Unexpected object type %s" % obj_type
                    )
                run = LocalMetadataProvider.get_object(
                    "run", "self", {}, None, *args[:RUN_ORDER]  # *[flow_id, run_id]
                )
                if not run:
                    raise MetaflowInternalError(
                        msg="Could not find run %s" % str(args[:RUN_ORDER])
                    )

                obj["tags"] = run.get("tags", [])
                obj["system_tags"] = run.get("system_tags", [])
                return obj
            return None

        if sub_type == "artifact":
            if obj_type not in ("root", "flow", "run", "step", "task"):
                raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)

            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            result = []
            if meta_path is None:
                return result

            successful_attempt = attempt
            if successful_attempt is None:
                attempt_done_files = os.path.join(meta_path, "sysmeta_attempt-done_*")
                attempts_done = sorted(glob.iglob(attempt_done_files))
                if attempts_done:
                    successful_attempt = int(
                        LocalMetadataProvider._read_json_file(attempts_done[-1])[
                            "value"
                        ]
                    )
            if successful_attempt is not None:
                which_artifact = "*"
                if len(args) >= sub_order:
                    which_artifact = args[sub_order - 1]
                artifact_files = os.path.join(
                    meta_path,
                    "%d_artifact_%s.json" % (successful_attempt, which_artifact),
                )
                for obj in glob.iglob(artifact_files):
                    result.append(LocalMetadataProvider._read_json_file(obj))

            # We are getting artifacts. We should overlay with ancestral run's tags
            run = LocalMetadataProvider.get_object(
                "run", "self", {}, None, *args[:RUN_ORDER]  # *[flow_id, run_id]
            )
            if not run:
                raise MetaflowInternalError(
                    msg="Could not find run %s" % str(args[:RUN_ORDER])
                )
            for obj in result:
                obj["tags"] = run.get("tags", [])
                obj["system_tags"] = run.get("system_tags", [])

            if len(result) == 1:
                return result[0]
            return result

        if sub_type == "metadata":
            # artifact is not expected because if obj_type=artifact on function entry, we transform to =task
            if obj_type not in ("root", "flow", "run", "step", "task"):
                raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)
            result = []
            meta_path = LocalMetadataProvider._get_metadir(*args[:obj_order])
            if meta_path is None:
                return result
            files = os.path.join(meta_path, "sysmeta_*")
            for obj in glob.iglob(files):
                result.append(LocalMetadataProvider._read_json_file(obj))
            return result

        # For the other types, we locate all the objects we need to find and return them
        if obj_type not in ("root", "flow", "run", "step", "task"):
            raise MetaflowInternalError(msg="Unexpected object type %s" % obj_type)
        if sub_type not in ("flow", "run", "step", "task"):
            raise MetaflowInternalError(msg="unexpected sub type %s" % sub_type)
        obj_path = LocalMetadataProvider._make_path(
            *args[:obj_order], create_on_absent=False
        )
        result = []
        if obj_path is None:
            return result
        skip_dirs = "*/" * (sub_order - obj_order)
        all_meta = os.path.join(obj_path, skip_dirs, LocalStorage.METADATA_DIR)
        SelfInfo = collections.namedtuple("SelfInfo", ["filepath", "run_id"])
        self_infos = []
        for meta_path in glob.iglob(all_meta):
            self_file = os.path.join(meta_path, "_self.json")
            if not os.path.isfile(self_file):
                continue
            run_id = None
            # flow and run do not need info from ancestral run
            if sub_type in ("step", "task"):
                run_id = LocalMetadataProvider._deduce_run_id_from_meta_dir(
                    meta_path, sub_type
                )
                # obj_type IS run, or more granular than run, let's do sanity check vs args
                if obj_order >= RUN_ORDER:
                    if run_id != args[RUN_ORDER - 1]:
                        raise MetaflowInternalError(
                            msg="Unexpected run id %s deduced from meta path" % run_id
                        )
            self_infos.append(SelfInfo(filepath=self_file, run_id=run_id))

        for self_info in self_infos:
            obj = LocalMetadataProvider._read_json_file(self_info.filepath)
            if self_info.run_id:
                flow_id_from_args = args[0]
                run = LocalMetadataProvider.get_object(
                    "run",
                    "self",
                    {},
                    None,
                    flow_id_from_args,
                    self_info.run_id,
                )
                if not run:
                    raise MetaflowInternalError(
                        msg="Could not find run %s, %s"
                        % (flow_id_from_args, self_info.run_id)
                    )
                obj["tags"] = run.get("tags", [])
                obj["system_tags"] = run.get("system_tags", [])
            result.append(obj)

        return MetadataProvider._apply_filter(result, filters)

    @staticmethod
    def _deduce_run_id_from_meta_dir(meta_dir_path, sub_type):
        curr_order = ObjectOrder.type_to_order(sub_type)
        levels_to_ascend = curr_order - ObjectOrder.type_to_order("run")
        if levels_to_ascend < 0:
            return None
        curr_path = meta_dir_path
        for _ in range(levels_to_ascend + 1):  # +1 to account for ../_meta
            curr_path, _ = os.path.split(curr_path)
        _, run_id = os.path.split(curr_path)
        if not run_id:
            raise MetaflowInternalError(
                "Failed to deduce run_id from meta dir %s" % meta_dir_path
            )
        return run_id

    @staticmethod
    def _makedirs(path):
        # this is for python2 compatibility.
        # Python3 has os.makedirs(exist_ok=True).
        try:
            os.makedirs(path)
        except OSError as x:
            if x.errno == 17:
                # Error raised when directory exists
                return
            else:
                raise

    @staticmethod
    def _persist_tags_for_run(flow_id, run_id, tags, system_tags):
        subpath = LocalMetadataProvider._create_and_get_metadir(
            flow_name=flow_id, run_id=run_id
        )
        selfname = os.path.join(subpath, "_self.json")
        if not os.path.isfile(selfname):
            raise MetaflowInternalError(
                msg="Could not verify Run existence on disk - missing %s" % selfname
            )
        LocalMetadataProvider._save_meta(
            subpath,
            {
                "_self": MetadataProvider._run_to_json_static(
                    flow_id, run_id=run_id, tags=tags, sys_tags=system_tags
                )
            },
            allow_overwrite=True,
        )

    def _ensure_meta(
        self, obj_type, run_id, step_name, task_id, tags=None, sys_tags=None
    ):
        if tags is None:
            tags = set()
        if sys_tags is None:
            sys_tags = set()
        subpath = self._create_and_get_metadir(
            self._flow_name, run_id, step_name, task_id
        )
        selfname = os.path.join(subpath, "_self.json")
        self._makedirs(subpath)
        if os.path.isfile(selfname):
            # There is a race here, but we are not aiming to make this as solid as
            # the metadata service. This is used primarily for concurrent resumes,
            # so it is highly unlikely that this combination (multiple resumes of
            # the same flow on the same machine) happens.
            return False
        # In this case the metadata information does not exist, so we create it
        self._save_meta(
            subpath,
            {
                "_self": self._object_to_json(
                    obj_type,
                    run_id,
                    step_name,
                    task_id,
                    self.sticky_tags.union(tags),
                    self.sticky_sys_tags.union(sys_tags),
                )
            },
        )
        return True

    def _new_run(self, run_id, tags=None, sys_tags=None):
        self._ensure_meta("flow", None, None, None)
        return self._ensure_meta("run", run_id, None, None, tags, sys_tags)

    def _new_task(
        self, run_id, step_name, task_id, attempt=0, tags=None, sys_tags=None
    ):
        self._ensure_meta("step", run_id, step_name, None)
        to_return = self._ensure_meta(
            "task", run_id, step_name, task_id, tags, sys_tags
        )
        self._register_system_metadata(run_id, step_name, task_id, attempt)
        return to_return

    @staticmethod
    def _make_path(
        flow_name=None, run_id=None, step_name=None, task_id=None, create_on_absent=True
    ):

        from metaflow.plugins.datastores.local_storage import LocalStorage

        if LocalStorage.datastore_root is None:

            def print_clean(line, **kwargs):
                print(line)

            LocalStorage.datastore_root = LocalStorage.get_datastore_root_from_config(
                print_clean, create_on_absent=create_on_absent
            )
        if LocalStorage.datastore_root is None:
            return None

        if flow_name is None:
            return LocalStorage.datastore_root
        components = []
        if flow_name:
            components.append(flow_name)
            if run_id:
                components.append(run_id)
                if step_name:
                    components.append(step_name)
                    if task_id:
                        components.append(task_id)
        return LocalStorage().full_uri(LocalStorage.path_join(*components))

    @staticmethod
    def _create_and_get_metadir(
        flow_name=None, run_id=None, step_name=None, task_id=None
    ):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        root_path = LocalMetadataProvider._make_path(
            flow_name, run_id, step_name, task_id
        )
        subpath = os.path.join(root_path, LocalStorage.METADATA_DIR)
        LocalMetadataProvider._makedirs(subpath)
        return subpath

    @staticmethod
    def _get_metadir(flow_name=None, run_id=None, step_name=None, task_id=None):
        from metaflow.plugins.datastores.local_storage import LocalStorage

        root_path = LocalMetadataProvider._make_path(
            flow_name, run_id, step_name, task_id, create_on_absent=False
        )
        if root_path is None:
            return None
        subpath = os.path.join(root_path, LocalStorage.METADATA_DIR)
        if os.path.isdir(subpath):
            return subpath
        return None

    @staticmethod
    def _dump_json_to_file(filepath, data, allow_overwrite=False):
        if os.path.isfile(filepath) and not allow_overwrite:
            return
        try:
            with tempfile.NamedTemporaryFile(
                mode="w", dir=os.path.dirname(filepath), delete=False
            ) as f:
                json.dump(data, f)
            os.rename(f.name, filepath)
        finally:
            # clean up in case anything goes wrong
            if f and os.path.isfile(f.name):
                os.remove(f.name)

    @staticmethod
    def _read_json_file(filepath):
        with open(filepath, "r") as f:
            return json.load(f)

    @staticmethod
    def _save_meta(root_dir, metadict, allow_overwrite=False):
        for name, datum in metadict.items():
            filename = os.path.join(root_dir, "%s.json" % name)
            LocalMetadataProvider._dump_json_to_file(
                filename, datum, allow_overwrite=allow_overwrite
            )



import os
import threading

_client_cache = dict()


def _get_cache_key():
    return os.getpid(), threading.get_ident()


def _get_gs_storage_client_default():
    cache_key = _get_cache_key()
    if cache_key not in _client_cache:
        from google.cloud import storage
        import google.auth

        credentials, project_id = google.auth.default(scopes=storage.Client.SCOPE)
        _client_cache[cache_key] = storage.Client(
            credentials=credentials, project=project_id
        )
    return _client_cache[cache_key]


class GcpDefaultClientProvider(object):
    name = "gcp-default"

    @staticmethod
    def get_gs_storage_client(*args, **kwargs):
        return _get_gs_storage_client_default()

    @staticmethod
    def get_credentials(scopes, *args, **kwargs):
        import google.auth

        return google.auth.default(scopes=scopes)


cached_provider_class = None


def get_gs_storage_client():
    global cached_provider_class
    if cached_provider_class is None:
        from metaflow.metaflow_config import DEFAULT_GCP_CLIENT_PROVIDER
        from metaflow.plugins import GCP_CLIENT_PROVIDERS

        for p in GCP_CLIENT_PROVIDERS:
            if p.name == DEFAULT_GCP_CLIENT_PROVIDER:
                cached_provider_class = p
                break
        else:
            raise ValueError(
                "Cannot find GCP Client provider %s" % DEFAULT_GCP_CLIENT_PROVIDER
            )
    return cached_provider_class.get_gs_storage_client()


def get_credentials(scopes, *args, **kwargs):
    global cached_provider_class
    if cached_provider_class is None:
        from metaflow.metaflow_config import DEFAULT_GCP_CLIENT_PROVIDER
        from metaflow.plugins import GCP_CLIENT_PROVIDERS

        for p in GCP_CLIENT_PROVIDERS:
            if p.name == DEFAULT_GCP_CLIENT_PROVIDER:
                cached_provider_class = p
                break
        else:
            raise ValueError(
                "Cannot find GCP Client provider %s" % DEFAULT_GCP_CLIENT_PROVIDER
            )
    return cached_provider_class.get_credentials(scopes, *args, **kwargs)



import traceback
import json

# from .card import MetaflowCardComponent
from .basic import SerializationErrorComponent


def _render_component_safely(
    component, render_func, return_error_component, *args, **kwargs
):
    rendered_obj = None
    try:
        rendered_obj = render_func(component, *args, **kwargs)
    except:
        if not return_error_component:
            return None
        error_str = traceback.format_exc()
        rendered_obj = SerializationErrorComponent(
            component.__class__.__name__, error_str
        ).render()
    else:
        if not (type(rendered_obj) == str or type(rendered_obj) == dict):
            rendered_obj = SerializationErrorComponent(
                component.__class__.__name__,
                "Component render didn't return a `string` or `dict`",
            ).render()
        else:
            try:  # check if rendered object is json serializable.
                json.dumps(rendered_obj)
            except (TypeError, OverflowError) as e:
                rendered_obj = SerializationErrorComponent(
                    component.__class__.__name__,
                    "Rendered Component cannot be JSON serialized. \n\n %s" % str(e),
                ).render()
    return rendered_obj


def render_safely(func):
    """
    This is a decorator that can be added to any `MetaflowCardComponent.render`
    The goal is to render subcomponents safely and ensure that they are JSON serializable.
    """
    # expects a renderer func
    def ret_func(self, *args, **kwargs):
        return _render_component_safely(self, func, True, *args, **kwargs)

    return ret_func



from io import BytesIO
from .s3util import aws_retry, get_s3_client

try:
    # python2
    from urlparse import urlparse
except:
    # python3
    from urllib.parse import urlparse


class S3Tail(object):
    def __init__(self, s3url):
        url = urlparse(s3url)
        self.s3, self.ClientError = get_s3_client()
        self._bucket = url.netloc
        self._key = url.path.lstrip("/")
        self._pos = 0
        self._tail = b""

    def reset_client(self, hard_reset=False):
        # This method is required by @aws_retry
        if hard_reset or self.s3 is None:
            self.s3, self.ClientError = get_s3_client()

    def clone(self, s3url):
        tail = S3Tail(s3url)
        tail._pos = self._pos
        tail._tail = self._tail
        return tail

    @property
    def bytes_read(self):
        return self._pos

    @property
    def tail(self):
        return self._tail

    def __iter__(self):
        buf = self._fill_buf()
        if buf is not None:
            for line in buf:
                if line.endswith(b"\n"):
                    yield line
                else:
                    self._tail = line
                    break

    @aws_retry
    def _make_range_request(self):
        try:
            return self.s3.get_object(
                Bucket=self._bucket, Key=self._key, Range="bytes=%d-" % self._pos
            )
        except self.ClientError as err:
            code = err.response["Error"]["Code"]
            # NOTE we deliberately regard NoSuchKey as an ignorable error.
            # We assume that the file just hasn't appeared in S3 yet.
            # Some S3 compatible storage systems like Dell EMC-ECS return 416 in-lieu
            # of InvalidRange - https://www.delltechnologies.com/asset/en-us/products/storage/technical-support/docu95766.pdf
            if code in ("InvalidRange", "NoSuchKey", "416"):
                return None
            else:
                raise

    def _fill_buf(self):
        resp = self._make_range_request()
        if resp is None:
            return None
        code = str(resp["ResponseMetadata"]["HTTPStatusCode"])
        if code[0] == "2":
            data = resp["Body"].read()
            if data:
                buf = BytesIO(self._tail + data)
                self._pos += len(data)
                self._tail = b""
                return buf
            else:
                return None
        elif code[0] == "5":
            return None
        else:
            raise Exception(
                "Retrieving %s/%s failed: %s" % (self._bucket, self._key, code)
            )



# Read an AWS source in a chunked manner.
# We read in chunks (at most 2GB -- here this is passed via max_chunk_size)
# because of https://bugs.python.org/issue42853 (Py3 bug); this also helps
# keep memory consumption lower
# NOTE: For some weird reason, if you pass a large value to
# read it delays the call, so we always pass it either what
# remains or 2GB, whichever is smallest.
def read_in_chunks(dst, src, src_sz, max_chunk_size):
    remaining = src_sz
    while remaining > 0:
        buf = src.read(min(remaining, max_chunk_size))
        # Py2 doesn't return the number of bytes written so calculate size
        # separately
        dst.write(buf)
        remaining -= len(buf)


from .local import MetaflowLocalNotFound, MetaflowLocalURLException, Local
from .s3 import MetaflowS3Exception, S3

# Import any additional datatools defined by a Metaflow extensions package
try:
    from metaflow.extension_support import get_modules, multiload_all

    multiload_all(get_modules("plugins.datatools"), "plugins.datatools", globals())
finally:
    # Erase all temporary names to avoid leaking things
    for _n in ["get_modules", "multiload_all"]:
        try:
            del globals()[_n]
        except KeyError:
            pass
    del globals()["_n"]



import base64
import json
from json import JSONDecodeError


from metaflow.exception import MetaflowException
from metaflow.plugins.secrets import SecretsProvider
import re
from metaflow.plugins.gcp.gs_storage_client_factory import get_credentials
from metaflow.metaflow_config import GCP_SECRET_MANAGER_PREFIX


class MetaflowGcpSecretsManagerBadResponse(MetaflowException):
    """Raised when the response from GCP Secrets Manager is not valid in some way"""


class MetaflowGcpSecretsManagerDuplicateKey(MetaflowException):
    """Raised when the response from GCP Secrets Manager contains duplicate keys"""


class MetaflowGcpSecretsManagerJSONParseError(MetaflowException):
    """Raised when the SecretString response from GCP Secrets Manager is not valid JSON"""


class MetaflowGcpSecretsManagerNotJSONObject(MetaflowException):
    """Raised when the SecretString response from GCP Secrets Manager is not valid JSON dictionary"""


def _sanitize_key_as_env_var(key):
    """
    Sanitize a key as an environment variable name.
    This is purely a convenience trade-off to cover common cases well, vs. introducing
    ambiguities (e.g. did the final '_' come from '.', or '-' or is original?).

    1/27/2023(jackie):

    We start with few rules and should *sparingly* add more over time.
    Also, it's TBD whether all possible providers will share the same sanitization logic.
    Therefore we will keep this function private for now
    """
    return key.replace("-", "_").replace(".", "_").replace("/", "_")


class GcpSecretManagerSecretsProvider(SecretsProvider):
    TYPE = "gcp-secret-manager"

    def get_secret_as_dict(self, secret_id, options={}, role=None):
        """
        Reads a secret from GCP Secrets Manager and returns it as a dictionary of environment variables.

        If the secret contains a string payload ("SecretString"):
        - if the `json` option is True:
            Secret will be parsed as a JSON. If successfully parsed, AND the JSON contains a
            top-level object, each entry K/V in the object will also be converted to an entry in the result. V will
            always be casted to a string (if not already a string).
        - If `json` option is False (default):
            Will be returned as a single entry in the result, with the key being the last part after / in secret_id.

        On GCP Secrets Manager, the secret payload is a binary blob. However, by default we interpret it as UTF8 encoded
        string. To disable this, set the `binary` option to True, the binary will be base64 encoded in the result.

        All keys in the result are sanitized to be more valid environment variable names. This is done on a best effort
        basis. Further validation is expected to be done by the invoking @secrets decorator itself.

        :param secret_id: GCP Secrets Manager secret ID
        :param options: unused
        :return: dict of environment variables. All keys and values are strings.
        """
        from google.cloud.secretmanager_v1.services.secret_manager_service import (
            SecretManagerServiceClient,
        )
        from google.cloud.secretmanager_v1.services.secret_manager_service.transports import (
            SecretManagerServiceTransport,
        )

        # Full secret id looks like projects/1234567890/secrets/mysecret/versions/latest
        #
        # We allow these forms of secret_id:
        #
        # 1. Full path like projects/1234567890/secrets/mysecret/versions/latest
        #    This is what you'd specify if you used to GCP SDK.
        #
        # 2. Full path but without the version like projects/1234567890/secrets/mysecret.
        #    This is what you see in the GCP console, makes it easier to copy & paste.
        #
        # 3. Simple string like mysecret
        #
        # 4. Simple string with /versions/<version> suffix like mysecret/versions/1

        # The latter two forms require METAFLOW_GCP_SECRET_MANAGER_PREFIX to be set.

        match_full = re.match(r"^projects/\d+/secrets/([\w\-]+)(/versions/([\w\-]+))?$", secret_id)
        match_partial = re.match(r"^([\w\-]+)(/versions/[\w\-]+)?$", secret_id)
        if match_full:
            # Full path
            env_var_name = match_full.group(1)
            if match_full.group(3):
                # With version specified
                full_secret_name = secret_id
            else:
                # No version specified, use latest
                full_secret_name = secret_id + "/versions/latest"
        elif match_partial:
            # Partial path, possibly with /versions/<version> suffix
            env_var_name = secret_id
            if not GCP_SECRET_MANAGER_PREFIX:
                raise ValueError(
                    f"Cannot use simple secret_id without setting METAFLOW_GCP_SECRET_MANAGER_PREFIX. {GCP_SECRET_MANAGER_PREFIX}"
                )
            if match_partial.group(2):
                # With version specified
                full_secret_name = f"{GCP_SECRET_MANAGER_PREFIX}{secret_id}"
                env_var_name = match_partial.group(1)
            else:
                # No version specified, use latest
                full_secret_name = (
                    f"{GCP_SECRET_MANAGER_PREFIX}{secret_id}/versions/latest"
                )
        else:
            raise ValueError(
                f"Invalid secret_id: {secret_id}. Must be either a full path or a simple string."
            )

        result = {}

        def _sanitize_and_add_entry_to_result(k, v):
            # Two jobs - sanitize, and check for dupes
            sanitized_k = _sanitize_key_as_env_var(k)
            if sanitized_k in result:
                raise MetaflowGcpSecretsManagerDuplicateKey(
                    "Duplicate key in secret: '%s' (sanitizes to '%s')"
                    % (k, sanitized_k)
                )
            result[sanitized_k] = v

        credentials, _ = get_credentials(
            scopes=SecretManagerServiceTransport.AUTH_SCOPES
        )
        client = SecretManagerServiceClient(credentials=credentials)
        response = client.access_secret_version(request={"name": full_secret_name})
        payload_str = response.payload.data.decode("UTF-8")
        if options.get("json", False):
            obj = json.loads(payload_str)
            if type(obj) == dict:
                for k, v in obj.items():
                    # We try to make it work here - cast to string always
                    _sanitize_and_add_entry_to_result(k, str(v))
            else:
                raise MetaflowGcpSecretsManagerNotJSONObject(
                    "Secret string is a JSON, but not an object (dict-like) - actual type %s."
                    % type(obj)
                )
        else:
            if options.get("env_var_name"):
                env_var_name = options["env_var_name"]

            if options.get("binary", False):
                _sanitize_and_add_entry_to_result(
                    env_var_name, base64.b64encode(response.payload.data)
                )
            else:
                _sanitize_and_add_entry_to_result(env_var_name, payload_str)

        return result



from .base_sensor import AirflowSensorDecorator
from ..airflow_utils import SensorNames
from ..exception import AirflowException


class S3KeySensorDecorator(AirflowSensorDecorator):
    """
    The `@airflow_s3_key_sensor` decorator attaches a Airflow [S3KeySensor](https://airflow.apache.org/docs/apache-airflow-providers-amazon/stable/_api/airflow/providers/amazon/aws/sensors/s3/index.html#airflow.providers.amazon.aws.sensors.s3.S3KeySensor)
    before the start step of the flow. This decorator only works when a flow is scheduled on Airflow
    and is compiled using `airflow create`. More than one `@airflow_s3_key_sensor` can be
    added as a flow decorators. Adding more than one decorator will ensure that `start` step
    starts only after all sensors finish.

    Parameters
    ----------
    timeout : int
        Time, in seconds before the task times out and fails. (Default: 3600)
    poke_interval : int
        Time in seconds that the job should wait in between each try. (Default: 60)
    mode : str
        How the sensor operates. Options are: { poke | reschedule }. (Default: "poke")
    exponential_backoff : bool
        allow progressive longer waits between pokes by using exponential backoff algorithm. (Default: True)
    pool : str
        the slot pool this task should run in,
        slot pools are a way to limit concurrency for certain tasks. (Default:None)
    soft_fail : bool
        Set to true to mark the task as SKIPPED on failure. (Default: False)
    name : str
        Name of the sensor on Airflow
    description : str
        Description of sensor in the Airflow UI
    bucket_key : Union[str, List[str]]
        The key(s) being waited on. Supports full s3:// style url or relative path from root level.
        When it's specified as a full s3:// url, please leave `bucket_name` as None
    bucket_name : str
        Name of the S3 bucket. Only needed when bucket_key is not provided as a full s3:// url.
        When specified, all the keys passed to bucket_key refers to this bucket. (Default:None)
    wildcard_match : bool
        whether the bucket_key should be interpreted as a Unix wildcard pattern. (Default: False)
    aws_conn_id : str
        a reference to the s3 connection on Airflow. (Default: None)
    verify : bool
        Whether or not to verify SSL certificates for S3 connection. (Default: None)
    """

    name = "airflow_s3_key_sensor"
    operator_type = SensorNames.S3_SENSOR
    # Arg specification can be found here :
    # https://airflow.apache.org/docs/apache-airflow-providers-amazon/stable/_api/airflow/providers/amazon/aws/sensors/s3/index.html#airflow.providers.amazon.aws.sensors.s3.S3KeySensor
    defaults = dict(
        **AirflowSensorDecorator.defaults,
        bucket_key=None,  # Required
        bucket_name=None,
        wildcard_match=False,
        aws_conn_id=None,
        verify=None,  # `verify (Optional[Union[str, bool]])` Whether or not to verify SSL certificates for S3 connection.
        #  `verify` is a airflow variable.
    )

    def validate(self):
        if self.attributes["bucket_key"] is None:
            raise AirflowException(
                "`bucket_key` for `@%s`cannot be empty." % (self.name)
            )
        super().validate()



from setuptools import setup, find_packages

with open("metaflow/version.py", mode="r") as f:
    version = f.read().splitlines()[0].split("=")[1].strip(" \"'")

setup(
    include_package_data=True,
    name="ob-metaflow",
    version=version,
    description="Metaflow: More Data Science, Less Engineering",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    author="Netflix, Outerbounds & the Metaflow Community",
    author_email="help@outerbounds.co",
    license="Apache License 2.0",
    packages=find_packages(exclude=["metaflow_test"]),
    py_modules=[
        "metaflow",
    ],
    package_data={
        "metaflow": [
            "tutorials/*/*",
            "plugins/env_escape/configurations/*/*",
            "py.typed",
            "**/*.pyi",
        ]
    },
    entry_points="""
        [console_scripts]
        metaflow=metaflow.cmd.main_cli:start
      """,
    install_requires=[
        "requests",
        "boto3",
        "pylint",
        "kubernetes",
    ],
    extras_require={
        "stubs": ["ob-metaflow-stubs==%s" % version],
    },
)



import functools
import pickle
from typing import Any

from .consts import (
    OP_GETATTR,
    OP_SETATTR,
    OP_DELATTR,
    OP_CALL,
    OP_CALLATTR,
    OP_CALLONCLASS,
    OP_DEL,
    OP_REPR,
    OP_STR,
    OP_HASH,
    OP_PICKLE,
    OP_DIR,
    OP_INIT,
    OP_SUBCLASSCHECK,
)

from .exception_transferer import ExceptionMetaClass

DELETED_ATTRS = frozenset(["__array_struct__", "__array_interface__"])

# These attributes are accessed directly on the stub (not directly forwarded)
LOCAL_ATTRS = (
    frozenset(
        [
            "___remote_class_name___",
            "___identifier___",
            "___connection___",
            "___local_overrides___",
            "___is_returned_exception___",
            "___exception_attributes___",
            "__class__",
            "__init__",
            "__del__",
            "__delattr__",
            "__dir__",
            "__doc__",
            "__getattr__",
            "__getattribute__",
            "__hash__",
            "__instancecheck__",
            "__subclasscheck__",
            "__init__",
            "__metaclass__",
            "__module__",
            "__name__",
            "__new__",
            "__reduce__",
            "__reduce_ex__",
            "__repr__",
            "__setattr__",
            "__slots__",
            "__str__",
            "__weakref__",
            "__dict__",
            "__methods__",
            "__exit__",
        ]
    )
    | DELETED_ATTRS
)

NORMAL_METHOD = 0
STATIC_METHOD = 1
CLASS_METHOD = 2


def fwd_request(stub, request_type, *args, **kwargs):
    connection = object.__getattribute__(stub, "___connection___")
    if connection:
        return connection.stub_request(stub, request_type, *args, **kwargs)
    raise RuntimeError(
        "Returned exception stub cannot be used to make further remote requests"
    )


class StubMetaClass(type):
    def __repr__(cls):
        if cls.__module__:
            return "<stub class '%s.%s'>" % (cls.__module__, cls.__name__)
        else:
            return "<stub class '%s'>" % (cls.__name__,)


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""

    # Compatibility 2/3. Remove when only 3 support
    class metaclass(type):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

    return type.__new__(metaclass, "temporary_class", (), {})


class Stub(with_metaclass(StubMetaClass, object)):
    """
    Local reference to a remote object.

    The stub looks and behaves like the remote object but all operations on the stub
    happen on the remote side (server).
    """

    __slots__ = ()
    # def __iter__(self):  # FIXME: Keep debugger QUIET!!
    #    raise AttributeError

    def __init__(
        self, connection, remote_class_name, identifier, _is_returned_exception=False
    ):
        self.___remote_class_name___ = remote_class_name
        self.___identifier___ = identifier
        self.___connection___ = connection
        # If it is a returned exception (ie: it was raised by the server), it behaves
        # a bit differently for methods like __str__ and __repr__ (we try not to get
        # stuff from the server)
        self.___is_returned_exception___ = _is_returned_exception

    def __del__(self):
        try:
            if not self.___is_returned_exception___:
                fwd_request(self, OP_DEL)
        except Exception:
            # raised in a destructor, most likely on program termination,
            # when the connection might have already been closed.
            # it's safe to ignore all exceptions here
            pass

    def __getattribute__(self, name):
        if name in LOCAL_ATTRS:
            if name == "__doc__":
                return self.__getattr__("__doc__")
            elif name in DELETED_ATTRS:
                raise AttributeError()
            else:
                return object.__getattribute__(self, name)
        elif name == "__call__":  # IronPython issue #10
            return object.__getattribute__(self, "__call__")
        elif name == "__array__":
            return object.__getattribute__(self, "__array__")
        else:
            return object.__getattribute__(self, name)

    def __getattr__(self, name):
        if name in DELETED_ATTRS:
            raise AttributeError()
        return fwd_request(self, OP_GETATTR, name)

    def __delattr__(self, name):
        if name in LOCAL_ATTRS:
            object.__delattr__(self, name)
        else:
            if self.___is_returned_exception___:
                raise AttributeError()
            return fwd_request(self, OP_DELATTR, name)

    def __setattr__(self, name, value):
        if (
            name in LOCAL_ATTRS
            or name in self.___local_overrides___
            or self.___is_returned_exception___
        ):
            object.__setattr__(self, name, value)
        else:
            fwd_request(self, OP_SETATTR, name, value)

    def __dir__(self):
        return fwd_request(self, OP_DIR)

    def __hash__(self):
        return fwd_request(self, OP_HASH)

    def __repr__(self):
        if self.___is_returned_exception___:
            return self.__exception_repr__()
        return fwd_request(self, OP_REPR)

    def __str__(self):
        if self.___is_returned_exception___:
            return self.__exception_str__()
        return fwd_request(self, OP_STR)

    def __exit__(self, exc, typ, tb):
        raise NotImplementedError
        # FIXME
        # return fwd_request(self, OP_CTX_EXIT, exc)  # can't pass type nor traceback

    def __reduce_ex__(self, proto):
        # support for pickling
        return pickle.loads, (fwd_request(self, OP_PICKLE, proto),)

    @classmethod
    def __subclasshook__(cls, parent):
        if parent.__bases__[0] == Stub:
            raise NotImplementedError  # Follow the usual mechanism
        # If this is not a stub, we go over to the other side
        parent_name = "%s.%s" % (parent.__module__, parent.__name__)
        return cls.___class_connection___.stub_request(
            None, OP_SUBCLASSCHECK, cls.___class_remote_class_name___, parent_name, True
        )


def _make_method(method_type, connection, class_name, name, doc):
    if name == "__call__":

        def __call__(_self, *args, **kwargs):
            return fwd_request(_self, OP_CALL, *args, **kwargs)

        __call__.__doc__ = doc
        return __call__

    def method(_self, *args, **kwargs):
        return fwd_request(_self, OP_CALLATTR, name, *args, **kwargs)

    def static_method(connection, class_name, name, *args, **kwargs):
        return connection.stub_request(
            None, OP_CALLONCLASS, class_name, name, True, *args, **kwargs
        )

    def class_method(connection, class_name, name, cls, *args, **kwargs):
        return connection.stub_request(
            None, OP_CALLONCLASS, class_name, name, False, *args, **kwargs
        )

    if method_type == NORMAL_METHOD:
        m = method
        m.__doc__ = doc
        m.__name__ = name
        return m
    if method_type == STATIC_METHOD:
        m = functools.partial(static_method, connection, class_name, name)
        m.__doc__ = doc
        m.__name__ = name
        m = staticmethod(m)
        return m
    if method_type == CLASS_METHOD:
        m = functools.partial(class_method, connection, class_name, name)
        m.__doc__ = doc
        m.__name__ = name
        m = classmethod(m)
        return m


class MetaWithConnection(StubMetaClass):
    # The use of this metaclass is so that we can support two modes when
    # instantiating a subclass of Stub. Suppose we have a class Foo which is a stub.
    # There are two ways Foo is initialized:
    #  - when it is returned from the remote side, in which case we do
    #    Foo(class_name, connection, identifier)
    #  - when it is created locally and needs to actually forward to __init__ on
    #    the remote side. In this case, we want the user to be able to do
    #    Foo(*args, **kwargs) (whatever the usual arguments to __init__ are)
    #
    # With this metaclass, we do just that. We introspect arguments and look to
    # see if the first one is the connection which would indicate that we are in
    # the first case. If that is the case, we just pass everything down to the
    # super __call__ and go our merry way. If this is not the case, we will
    # use the connection we saved when creating this metaclass and call
    # OP_INIT to create the object

    def __new__(cls, class_name, base_classes, class_dict, connection):
        return type.__new__(cls, class_name, base_classes, class_dict)

    def __init__(cls, class_name, base_classes, class_dict, connection):
        cls.___class_remote_class_name___ = class_name
        cls.___class_connection___ = connection
        super(MetaWithConnection, cls).__init__(class_name, base_classes, class_dict)

    def __call__(cls, *args, **kwargs):
        if len(args) > 0 and id(args[0]) == id(cls.___class_connection___):
            return super(MetaWithConnection, cls).__call__(*args, **kwargs)
        else:
            return cls.___class_connection___.stub_request(
                None, OP_INIT, cls.___class_remote_class_name___, *args, **kwargs
            )

    def __subclasscheck__(cls, subclass):
        subclass_name = "%s.%s" % (subclass.__module__, subclass.__name__)
        if subclass.__bases__[0] == Stub:
            subclass_name = subclass.___class_remote_class_name___
        return cls.___class_connection___.stub_request(
            None,
            OP_SUBCLASSCHECK,
            cls.___class_remote_class_name___,
            subclass_name,
        )

    def __instancecheck__(cls, instance):
        if type(instance) == cls:
            # Fast path if it's just an object of this class
            return True
        # Goes to __subclasscheck__ above
        return cls.__subclasscheck__(type(instance))


class MetaExceptionWithConnection(StubMetaClass, ExceptionMetaClass):
    def __new__(cls, class_name, base_classes, class_dict, connection):
        return type.__new__(cls, class_name, base_classes, class_dict)

    def __init__(cls, class_name, base_classes, class_dict, connection):
        cls.___class_remote_class_name___ = class_name
        cls.___class_connection___ = connection

        # We call the one on ExceptionMetaClass which does everything needed (StubMetaClass
        # does not do anything special for init)
        ExceptionMetaClass.__init__(cls, class_name, base_classes, class_dict)

        # Restore __str__ and __repr__ to the original ones because we need to determine
        # if we call them depending on whether or not the object is a returned exception
        # or not
        cls.__exception_str__ = cls.__str__
        cls.__exception_repr__ = cls.__repr__
        cls.__str__ = cls.__orig_str__
        cls.__repr__ = cls.__orig_repr__

    def __call__(cls, *args, **kwargs):
        # Very similar to the other case but we also need to be able to detect
        # local instantiation of an exception so that we can set the __is_returned_exception__
        if len(args) > 0 and id(args[0]) == id(cls.___class_connection___):
            return super(MetaExceptionWithConnection, cls).__call__(*args, **kwargs)
        elif kwargs and kwargs.get("_is_returned_exception", False):
            return super(MetaExceptionWithConnection, cls).__call__(
                None, None, None, _is_returned_exception=True
            )
        else:
            return cls.___class_connection___.stub_request(
                None, OP_INIT, cls.___class_remote_class_name___, *args, **kwargs
            )

    # The issue is that for a proxied object that is also an exception, we now have
    # two classes representing it, one that includes the Stub class and one that doesn't
    # Concretely:
    #  - test.MyException would return a class that derives from Stub
    #  - test.MySubException would return a class that derives from Stub and test.MyException
    #    but WITHOUT the Stub portion (see get_local_class).
    #  - we want issubclass(test.MySubException, test.MyException) to return True and
    #    the same with instance checks.
    def __instancecheck__(cls, instance):
        return cls.__subclasscheck__(type(instance))

    def __subclasscheck__(cls, subclass):
        # __mro__[0] is this class itself
        # __mro__[1] is the stub so we start checking at 2
        return any(
            [
                subclass.__mro__[i] in cls.__mro__[2:]
                for i in range(2, len(subclass.__mro__))
            ]
        )


def create_class(
    connection,
    class_name,
    overriden_methods,
    getattr_overrides,
    setattr_overrides,
    class_methods,
    parents,
):
    class_dict = {
        "__slots__": [
            "___remote_class_name___",
            "___identifier___",
            "___connection___",
            "___is_returned_exception___",
        ]
    }
    for name, doc in class_methods.items():
        method_type = NORMAL_METHOD
        if name.startswith("___s___"):
            name = name[7:]
            method_type = STATIC_METHOD
        elif name.startswith("___c___"):
            name = name[7:]
            method_type = CLASS_METHOD
        if name in overriden_methods:
            if method_type == NORMAL_METHOD:
                class_dict[name] = (
                    lambda override, orig_method: lambda obj, *args, **kwargs: override(
                        obj, functools.partial(orig_method, obj), *args, **kwargs
                    )
                )(
                    overriden_methods[name],
                    _make_method(method_type, connection, class_name, name, doc),
                )
            elif method_type == STATIC_METHOD:
                class_dict[name] = (
                    lambda override, orig_method: lambda *args, **kwargs: override(
                        orig_method, *args, **kwargs
                    )
                )(
                    overriden_methods[name],
                    _make_method(method_type, connection, class_name, name, doc),
                )
            elif method_type == CLASS_METHOD:
                class_dict[name] = (
                    lambda override, orig_method: lambda cls, *args, **kwargs: override(
                        cls, functools.partial(orig_method, cls), *args, **kwargs
                    )
                )(
                    overriden_methods[name],
                    _make_method(method_type, connection, class_name, name, doc),
                )
        elif name not in LOCAL_ATTRS:
            class_dict[name] = _make_method(
                method_type, connection, class_name, name, doc
            )
    # Check for any getattr/setattr overrides
    special_attributes = set(getattr_overrides.keys())
    special_attributes.update(set(setattr_overrides.keys()))
    overriden_attrs = set()
    for attr in special_attributes:
        getter = getattr_overrides.get(attr)
        setter = setattr_overrides.get(attr)
        if getter is not None:
            getter = lambda x, name=attr, inner=getter: inner(
                x, name, lambda y=x, name=name: y.__getattr__(name)
            )
        if setter is not None:
            setter = lambda x, value, name=attr, inner=setter: inner(
                x,
                name,
                lambda val, y=x, name=name: fwd_request(y, OP_SETATTR, name, val),
                value,
            )
            overriden_attrs.add(attr)
        class_dict[attr] = property(getter, setter)
    if parents:
        # This means this is also an exception so we add a few more things to it
        # so that it
        # This is copied from ExceptionMetaClass in exception_transferer.py
        for n in ("_exception_str", "_exception_repr", "_exception_tb"):
            class_dict[n] = property(
                lambda self, n=n: getattr(self, "%s_val" % n, "<missing>"),
                lambda self, v, n=n: setattr(self, "%s_val" % n, v),
            )

        def _do_str(self):
            text = self._exception_str
            text += "\n\n===== Remote (on server) traceback =====\n"
            text += self._exception_tb
            text += "========================================\n"
            return text

        class_dict["__exception_str__"] = _do_str
        class_dict["__exception_repr__"] = lambda self: self._exception_repr
    else:
        # If we are based on an exception, we already have __weakref__ so we don't add
        # it but not the case if we are not.
        class_dict["__slots__"].append("__weakref__")

    class_module, class_name_only = class_name.rsplit(".", 1)
    class_dict["___local_overrides___"] = overriden_attrs
    class_dict["__module__"] = class_module
    if parents:
        to_return = MetaExceptionWithConnection(
            class_name, (Stub, *parents), class_dict, connection
        )
    else:
        to_return = MetaWithConnection(class_name, (Stub,), class_dict, connection)
    to_return.__name__ = class_name_only
    return to_return
